"""
Django Helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008 Jutda. Copyright 2018 Timothy Hobbs. All Rights Reserved.
See LICENSE for details.
"""

# import base64
from bs4 import BeautifulSoup
from datetime import timedelta
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import SimpleUploadedFile
from django.db.models import Q
from django.utils import encoding, timezone
from django.utils.translation import gettext as _
import email
from email import policy
from email.message import EmailMessage, MIMEPart
from email.utils import getaddresses
from email_reply_parser import EmailReplyParser
from helpdesk import settings as helpdesk_settings
from helpdesk.exceptions import DeleteIgnoredTicketException, IgnoreTicketException
from helpdesk.lib import process_attachments, safe_template_context
from helpdesk.models import FollowUp, IgnoreEmail, Queue, Ticket
from helpdesk.signals import new_ticket_done, update_ticket_done
import imaplib
import logging
import mimetypes
import oauthlib.oauth2 as oauth2lib
import os
from os.path import isfile, join
import poplib
import re
import requests_oauthlib
import socket
import ssl
import sys
from time import ctime
import traceback
import typing
from typing import List


# import User model, which may be a custom model
User = get_user_model()

MULTIPLE_USERS_SAME_EMAIL_MSG = "CC email address is linked to more than 1 active user"
STRIPPED_SUBJECT_STRINGS = [
    "Re: ",
    "Fw: ",
    "RE: ",
    "FW: ",
    "Automatic reply: ",
]

# Allow a custom default attached email name for the HTML formatted email if one is found
HTML_EMAIL_ATTACHMENT_FILENAME = _("email_html_body.html")


def process_email(quiet: bool = False, debug_to_stdout: bool = False):
    if debug_to_stdout:
        print("Extracting email into queues...")
    q: Queue()  # Typing ahead of time for loop to make it more useful in an IDE
    for q in Queue.objects.filter(
        email_box_type__isnull=False, allow_email_submission=True
    ):
        log_msg = f"Processing queue: {q.slug} Email address: {q.email_address}..."
        if debug_to_stdout:
            print(log_msg)
        logger = logging.getLogger("django.helpdesk.queue." + q.slug)
        logging_types = {
            "info": logging.INFO,
            "warn": logging.WARN,
            "error": logging.ERROR,
            "crit": logging.CRITICAL,
            "debug": logging.DEBUG,
        }
        if q.logging_type in logging_types:
            logger.setLevel(logging_types[q.logging_type])
        elif not q.logging_type or q.logging_type == "none":
            # disable all handlers so messages go to nowhere
            logger.handlers = []
            logger.propagate = False
        if quiet:
            logger.propagate = (
                False  # do not propagate to root logger that would log to console
            )
        # Log messages to specific file only if the queue has it configured
        if (
            q.logging_type in logging_types
        ) and q.logging_dir:  # if it's enabled and the dir is set
            log_file_handler = logging.FileHandler(
                join(q.logging_dir, q.slug + "_get_email.log")
            )
            logger.addHandler(log_file_handler)
        else:
            log_file_handler = None
            if not q.email_box_last_check:
                q.email_box_last_check = timezone.now() - timedelta(minutes=30)
        try:
            queue_time_delta = timedelta(minutes=q.email_box_interval or 0)
            if (q.email_box_last_check + queue_time_delta) < timezone.now():
                process_queue(q, logger=logger)
                q.email_box_last_check = timezone.now()
                q.save()
                log_msg: str = f"Queue successfully processed: {q.slug}"
                logger.info(log_msg)
                if debug_to_stdout:
                    print(log_msg)
        except Exception as e:
            logger.error(f"Queue processing failed: {q.slug} -- {e}", exc_info=True)
            if debug_to_stdout:
                print(f"Queue processing failed: {q.slug}")
                print("-" * 60)
                traceback.print_exc(file=sys.stdout)
        finally:
            # we must close the file handler correctly if it's created
            try:
                if log_file_handler:
                    log_file_handler.close()
            except Exception as e:
                logging.exception(e)
            try:
                if log_file_handler:
                    logger.removeHandler(log_file_handler)
            except Exception as e:
                logging.exception(e)
    if debug_to_stdout:
        print("Email extraction into queues completed.")


def pop3_sync(q, logger, server):
    server.getwelcome()
    try:
        server.stls()
    except Exception:
        logger.warning(
            "POP3 StartTLS failed or unsupported. Connection will be unencrypted."
        )
    server.user(q.email_box_user or helpdesk_settings.QUEUE_EMAIL_BOX_USER)
    server.pass_(q.email_box_pass or helpdesk_settings.QUEUE_EMAIL_BOX_PASSWORD)

    messagesInfo = server.list()[1]
    logger.info("Received %d messages from POP3 server" % len(messagesInfo))

    for msgRaw in messagesInfo:
        if type(msgRaw) is bytes:
            try:
                msg = msgRaw.decode("utf-8")
            except UnicodeError:
                # if couldn't decode easily, just leave it raw
                msg = msgRaw
        else:
            # already a str
            msg = msgRaw
        msgNum = msg.split(" ")[0]
        logger.info("Processing message %s" % msgNum)

        raw_content = server.retr(msgNum)[1]
        if type(raw_content[0]) is bytes:
            full_message = "\n".join([elm.decode("utf-8") for elm in raw_content])
        else:
            full_message = encoding.force_str("\n".join(raw_content), errors="replace")
        try:
            ticket = extract_email_metadata(
                message=full_message, queue=q, logger=logger
            )
        except IgnoreTicketException:
            logger.warning(
                "Message %s was ignored and will be left on POP3 server" % msgNum
            )
        except DeleteIgnoredTicketException:
            logger.warning(
                "Message %s was ignored and deleted from POP3 server" % msgNum
            )
            server.dele(msgNum)
        else:
            if ticket:
                server.dele(msgNum)
                logger.info(
                    "Successfully processed message %s, deleted from POP3 server"
                    % msgNum
                )
            else:
                logger.warning(
                    "Message %s was not successfully processed, and will be left on POP3 server"
                    % msgNum
                )

    server.quit()


def imap_sync(q, logger, server):
    try:
        try:
            server.starttls()
        except Exception:
            logger.warning(
                "IMAP4 StartTLS unsupported or failed. Connection will be unencrypted."
            )
        server.login(
            q.email_box_user or helpdesk_settings.QUEUE_EMAIL_BOX_USER,
            q.email_box_pass or helpdesk_settings.QUEUE_EMAIL_BOX_PASSWORD,
        )
        server.select(q.email_box_imap_folder)
    except imaplib.IMAP4.abort:
        logger.error(
            "IMAP login failed. Check that the server is accessible and that "
            "the username and password are correct."
        )
        server.logout()
        sys.exit()
    except ssl.SSLError:
        logger.error(
            "IMAP login failed due to SSL error. This is often due to a timeout. "
            "Please check your connection and try again."
        )
        server.logout()
        sys.exit()

    try:
        data = server.search(None, "NOT", "DELETED")[1]
        if data:
            msgnums = data[0].split()
            logger.info("Received %d messages from IMAP server" % len(msgnums))
            for num in msgnums:
                logger.info("Processing message %s" % num)
                data = server.fetch(num, "(RFC822)")[1]
                full_message = encoding.force_str(data[0][1], errors="replace")
                try:
                    ticket = extract_email_metadata(
                        message=full_message, queue=q, logger=logger
                    )
                except IgnoreTicketException:
                    logger.warning(
                        "Message %s was ignored and will be left on IMAP server" % num
                    )
                except DeleteIgnoredTicketException:
                    server.store(num, "+FLAGS", "\\Deleted")
                    logger.warning(
                        "Message %s was ignored and deleted from IMAP server" % num
                    )
                except TypeError as te:
                    # Log the error with stacktrace to help identify what went wrong
                    logger.error(
                        f"Unexpected error processing message: {te}", exc_info=True
                    )
                else:
                    if ticket:
                        server.store(num, "+FLAGS", "\\Deleted")
                        logger.info(
                            "Successfully processed message %s, deleted from IMAP server"
                            % num
                        )
                    else:
                        logger.warning(
                            "Message %s was not successfully processed, and will be left on IMAP server"
                            % num
                        )
    except imaplib.IMAP4.error:
        logger.error(
            "IMAP retrieve failed. Is the folder '%s' spelled correctly, and does it exist on the server?",
            q.email_box_imap_folder,
        )

    server.expunge()
    server.close()
    server.logout()


def imap_oauth_sync(q, logger, server):
    """
    IMAP eMail server with OAUTH authentication.
    Only tested against O365 implementation

    Uses HELPDESK OAUTH Dict in Settings.

    """

    try:
        logger.debug("Start Mailbox polling via IMAP OAUTH")

        client = oauth2lib.BackendApplicationClient(
            client_id=helpdesk_settings.HELPDESK_OAUTH["client_id"],
            scope=helpdesk_settings.HELPDESK_OAUTH["scope"],
        )

        oauth = requests_oauthlib.OAuth2Session(client=client)
        token = oauth.fetch_token(
            token_url=helpdesk_settings.HELPDESK_OAUTH["token_url"],
            client_id=helpdesk_settings.HELPDESK_OAUTH["client_id"],
            client_secret=helpdesk_settings.HELPDESK_OAUTH["secret"],
            include_client_id=True,
        )

        server.debug = helpdesk_settings.HELPDESK_IMAP_DEBUG_LEVEL
        # TODO: Perhaps store the authentication string template externally? Settings? Queue Table?
        server.authenticate(
            "XOAUTH2",
            lambda x: f"user={q.email_box_user}\x01auth=Bearer {token['access_token']}\x01\x01".encode(),
        )
        # Select the Inbound Mailbox folder
        server.select(q.email_box_imap_folder)

    except imaplib.IMAP4.abort as e1:
        logger.error(f"IMAP authentication failed in OAUTH: {e1}", exc_info=True)
        server.logout()
        sys.exit()

    except ssl.SSLError as e2:
        logger.error(
            f"IMAP login failed due to SSL error. (This is often due to a timeout): {e2}",
            exc_info=True,
        )
        server.logout()
        sys.exit()

    try:
        data = server.search(None, "NOT", "DELETED")[1]
        if data:
            msgnums = data[0].split()
            logger.info(f"Found {len(msgnums)} message(s) on IMAP server")
            for num in msgnums:
                logger.info(f"Processing message {num}")
                data = server.fetch(num, "(RFC822)")[1]
                full_message = encoding.force_str(data[0][1], errors="replace")

                try:
                    ticket = extract_email_metadata(
                        message=full_message, queue=q, logger=logger
                    )

                except IgnoreTicketException as itex:
                    logger.warning(f"Message {num} was ignored. {itex}")

                except DeleteIgnoredTicketException:
                    server.store(num, "+FLAGS", "\\Deleted")
                    server.expunge()
                    logger.warning(
                        "Message %s was ignored and deleted from IMAP server" % num
                    )

                except TypeError as te:
                    # Log the error with stacktrace to help identify what went wrong
                    logger.error(
                        f"Unexpected error processing message: {te}", exc_info=True
                    )

                else:
                    if ticket:
                        server.store(num, "+FLAGS", "\\Deleted")
                        server.expunge()
                        logger.info(
                            "Successfully processed message %s, deleted from IMAP server"
                            % num
                        )
                    else:
                        logger.warning(
                            "Message %s was not successfully processed, and will be left on IMAP server"
                            % num
                        )

    except imaplib.IMAP4.error:
        logger.error(
            "IMAP retrieve failed. Is the folder '%s' spelled correctly, and does it exist on the server?",
            q.email_box_imap_folder,
        )
    # Purged Flagged Messages & Logout
    server.expunge()
    server.close()
    server.logout()


def process_queue(q, logger):
    logger.info(
        f"***** {ctime()}: Begin processing mail for django-helpdesk queue: {q.title}"
    )

    if q.socks_proxy_type and q.socks_proxy_host and q.socks_proxy_port:
        try:
            import socks
        except ImportError:
            no_socks_msg = (
                "Queue has been configured with proxy settings, "
                "but no socks library was installed. Try to "
                "install PySocks via PyPI."
            )
            logger.error(no_socks_msg)
            raise ImportError(no_socks_msg)

        proxy_type = {
            "socks4": socks.SOCKS4,
            "socks5": socks.SOCKS5,
        }.get(q.socks_proxy_type)

        socks.set_default_proxy(
            proxy_type=proxy_type, addr=q.socks_proxy_host, port=q.socks_proxy_port
        )
        socket.socket = socks.socksocket

    email_box_type = helpdesk_settings.QUEUE_EMAIL_BOX_TYPE or q.email_box_type

    mail_defaults = {
        "pop3": {
            "ssl": {
                "port": 995,
                "init": poplib.POP3_SSL,
            },
            "insecure": {
                "port": 110,
                "init": poplib.POP3,
            },
            "sync": pop3_sync,
        },
        "imap": {
            "ssl": {
                "port": 993,
                "init": imaplib.IMAP4_SSL,
            },
            "insecure": {
                "port": 143,
                "init": imaplib.IMAP4,
            },
            "sync": imap_sync,
        },
        "oauth": {
            "ssl": {
                "port": 993,
                "init": imaplib.IMAP4_SSL,
            },
            "insecure": {
                "port": 143,
                "init": imaplib.IMAP4,
            },
            "sync": imap_oauth_sync,
        },
    }
    if email_box_type in mail_defaults:
        encryption = "insecure"
        if q.email_box_ssl or helpdesk_settings.QUEUE_EMAIL_BOX_SSL:
            encryption = "ssl"
        if not q.email_box_port:
            q.email_box_port = mail_defaults[email_box_type][encryption]["port"]

        server = mail_defaults[email_box_type][encryption]["init"](
            q.email_box_host or helpdesk_settings.QUEUE_EMAIL_BOX_HOST,
            int(q.email_box_port),
        )
        logger.info("Attempting %s server login" % email_box_type.upper())
        mail_defaults[email_box_type]["sync"](q, logger, server)

    elif email_box_type == "local":
        mail_dir = q.email_box_local_dir or "/var/lib/mail/helpdesk/"
        mail = [
            join(mail_dir, f) for f in os.listdir(mail_dir) if isfile(join(mail_dir, f))
        ]
        logger.info("Found %d messages in local mailbox directory" % len(mail))

        logger.info("Found %d messages in local mailbox directory" % len(mail))
        for i, m in enumerate(mail, 1):
            logger.info("Processing message %d" % i)
            with open(m, "r") as f:
                full_message = encoding.force_str(f.read(), errors="replace")
                try:
                    ticket = extract_email_metadata(
                        message=full_message, queue=q, logger=logger
                    )
                except IgnoreTicketException:
                    logger.warning(
                        "Message %d was ignored and will be left in local directory", i
                    )
                except DeleteIgnoredTicketException:
                    os.unlink(m)
                    logger.warning(
                        "Message %d was ignored and deleted local directory", i
                    )
                else:
                    if ticket:
                        logger.info(
                            "Successfully processed message %d, ticket/comment created.",
                            i,
                        )
                        try:
                            # delete message file if ticket was successful
                            os.unlink(m)
                        except OSError as e:
                            logger.error("Unable to delete message %d (%s).", i, str(e))
                        else:
                            logger.info("Successfully deleted message %d.", i)
                    else:
                        logger.warning(
                            "Message %d was not successfully processed, and will be left in local directory",
                            i,
                        )


def decodeUnknown(charset, string):
    if string and not isinstance(string, str):
        if not charset:
            try:
                return str(string, encoding="utf-8", errors="replace")
            except UnicodeError:
                return str(string, encoding="iso8859-1", errors="replace")
        return str(string, encoding=charset, errors="replace")
    return string


def decode_mail_headers(string):
    decoded = email.header.decode_header(string)
    return " ".join(
        [
            str(msg, encoding=charset, errors="replace") if charset else str(msg)
            for msg, charset in decoded
        ]
    )


def is_autoreply(message):
    """
    Accepting message as something with .get(header_name) method
    Returns True if it's likely to be auto-reply or False otherwise
    So we don't start mail loops
    """
    any_if_this = [
        False
        if not message.get("Auto-Submitted")
        else message.get("Auto-Submitted").lower() != "no",
        True
        if message.get("X-Auto-Response-Suppress") in ("DR", "AutoReply", "All")
        else False,
        message.get("List-Id"),
        message.get("List-Unsubscribe"),
    ]
    return any(any_if_this)


def create_ticket_cc(ticket, cc_list, logger):
    if not cc_list:
        return []
    # Local import to deal with non-defined / circular reference problem

    new_ticket_ccs = []
    from helpdesk.views.staff import subscribe_to_ticket_updates, User

    for __, cced_email in cc_list:
        cced_email = cced_email.strip()
        if cced_email == ticket.queue.email_address:
            continue

        user_id = None

        user_list = User.objects.filter(email=cced_email, is_active=True).values_list(
            "id", flat=True
        )
        count = user_list.count()
        if count == 0:
            if getattr(
                helpdesk_settings, "LOG_WARN_WHEN_CC_EMAIL_NOT_LINKED_TO_A_USER", False
            ):
                logger.warning(
                    f"CC email address is not linked to an active user: {cced_email}"
                )
        elif count > 1:
            if getattr(
                helpdesk_settings,
                "LOG_WARN_WHEN_CC_EMAIL_LINKED_TO_MORE_THAN_1_USER",
                True,
            ):
                logger.warning(f"{MULTIPLE_USERS_SAME_EMAIL_MSG}: {cced_email}")
        else:
            user_id = user_list[0]

        try:
            ticket_cc = subscribe_to_ticket_updates(
                ticket=ticket, user_id=user_id, email=cced_email
            )
            new_ticket_ccs.append(ticket_cc)
        except ValidationError:
            pass

    return new_ticket_ccs


def create_object_from_email_message(message, ticket_id, payload, files, logger):
    ticket, previous_followup, new = None, None, False
    now = timezone.now()

    queue = payload["queue"]
    sender_email = payload["sender_email"]

    to_list = getaddresses(message.get_all("To", []))
    cc_list = getaddresses(message.get_all("Cc", []))

    message_id = message.get("Message-Id")
    in_reply_to = message.get("In-Reply-To")

    if message_id:
        message_id = message_id.strip()

    if in_reply_to:
        in_reply_to = in_reply_to.strip()

    if in_reply_to is not None:
        try:
            queryset = FollowUp.objects.filter(message_id=in_reply_to).order_by("-date")
            if queryset.count() > 0:
                previous_followup = queryset.first()
                ticket = previous_followup.ticket
        except FollowUp.DoesNotExist:
            pass  # play along. The header may be wrong

    if previous_followup is None and ticket_id is not None:
        try:
            ticket = Ticket.objects.get(id=ticket_id)
        except Ticket.DoesNotExist:
            ticket = None
        else:
            new = False
            # Check if the ticket has been merged to another ticket
            if ticket.merged_to:
                logger.info("Ticket has been merged to %s" % ticket.merged_to.ticket)
                # Use the ticket in which it was merged to for next operations
                ticket = ticket.merged_to
    # New issue, create a new <Ticket> instance
    if ticket is None:
        if not getattr(settings, "QUEUE_EMAIL_BOX_UPDATE_ONLY", False):
            ticket = Ticket.objects.create(
                title=payload["subject"],
                queue=queue,
                submitter_email=sender_email,
                created=now,
                description=payload["body"],
                priority=payload["priority"],
            )
            ticket.save()
            logger.debug("Created new ticket %s-%s" % (ticket.queue.slug, ticket.id))
            new = True
        else:
            # Possibly an email with no body but has an attachment
            logger.debug(
                "The QUEUE_EMAIL_BOX_UPDATE_ONLY setting is True so new ticket not created."
            )
            return None
    # Old issue being re-opened
    elif ticket.status == Ticket.CLOSED_STATUS:
        ticket.status = Ticket.REOPENED_STATUS
        ticket.save()

    f = FollowUp(
        ticket=ticket,
        title=_(
            "E-Mail Received from %(sender_email)s" % {"sender_email": sender_email}
        ),
        date=now,
        public=True,
        comment=payload.get("full_body", payload["body"]) or "",
        message_id=message_id,
    )

    if ticket.status == Ticket.REOPENED_STATUS:
        f.new_status = Ticket.REOPENED_STATUS
        f.title = _(
            "Ticket Re-Opened by E-Mail Received from %(sender_email)s"
            % {"sender_email": sender_email}
        )

    f.save()
    logger.debug("Created new FollowUp for Ticket")

    logger.info(
        "[%s-%s] %s"
        % (
            ticket.queue.slug,
            ticket.id,
            ticket.title,
        )
    )

    if helpdesk_settings.HELPDESK_ENABLE_ATTACHMENTS:
        try:
            attached = process_attachments(f, files)
        except ValidationError as e:
            logger.error(str(e))
        else:
            for att_file in attached:
                logger.info(
                    "Attachment '%s' (with size %s) successfully added to ticket from email.",
                    att_file[0],
                    att_file[1].size,
                )

    context = safe_template_context(ticket)

    new_ticket_ccs = []
    new_ticket_ccs.append(create_ticket_cc(ticket, to_list + cc_list, logger))

    autoreply = is_autoreply(message)
    if autoreply:
        logger.info(
            "Message seems to be auto-reply, not sending any emails back to the sender"
        )
    else:
        send_info_email(message_id, f, ticket, context, queue, new)
    if new:
        # emit signal when a new ticket is created
        new_ticket_done.send(sender="create_object_from_email_message", ticket=ticket)
    else:
        # emit signal with followup when the ticket is updated
        update_ticket_done.send(sender="create_object_from_email_message", followup=f)
    return ticket


def send_info_email(
    message_id: str, f: FollowUp, ticket: Ticket, context: dict, queue: dict, new: bool
):
    # send mail to appropriate people now depending on what objects
    # were created and who was CC'd
    # Add auto-reply headers because it's an auto-reply and we must
    extra_headers = {
        "In-Reply-To": message_id,
        "Auto-Submitted": "auto-replied",
        "X-Auto-Response-Suppress": "All",
        "Precedence": "auto_reply",
    }
    if new:
        ticket.send(
            {
                "submitter": ("newticket_submitter", context),
                "new_ticket_cc": ("newticket_cc", context),
                "ticket_cc": ("newticket_cc", context),
            },
            fail_silently=True,
            extra_headers=extra_headers,
        )
    else:
        context.update(comment=f.comment)
        ticket.send(
            {
                "submitter": ("updated_submitter", context),
                "assigned_to": ("updated_owner", context),
            },
            fail_silently=True,
            extra_headers=extra_headers,
        )
        if queue.enable_notifications_on_email_events:
            ticket.send(
                {"ticket_cc": ("updated_cc", context)},
                fail_silently=True,
                extra_headers=extra_headers,
            )


def get_ticket_id_from_subject_slug(
    queue_slug: str, subject: str, logger: logging.Logger
) -> typing.Optional[int]:
    """Get a ticket id from the subject string

    Performs a match on the subject using the queue_slug as reference,
    returning the ticket id if a match is found.
    """
    matchobj = re.match(r".*\[" + queue_slug + r"-(?P<id>\d+)\]", subject)
    ticket_id = None
    if matchobj:
        # This is a reply or forward.
        ticket_id = matchobj.group("id")
        logger.info("Matched tracking ID %s-%s" % (queue_slug, ticket_id))
    else:
        logger.info("No tracking ID matched.")
    return ticket_id


def add_file_if_always_save_incoming_email_message(files_, message: str) -> None:
    """When `settings.HELPDESK_ALWAYS_SAVE_INCOMING_EMAIL_MESSAGE` is `True`
    add a file to the files_ list"""
    if getattr(settings, "HELPDESK_ALWAYS_SAVE_INCOMING_EMAIL_MESSAGE", False):
        # save message as attachment in case of some complex markup renders
        # wrong
        files_.append(
            SimpleUploadedFile(
                _("original_message.eml").replace(
                    ".eml", timezone.localtime().strftime("_%d-%m-%Y_%H:%M") + ".eml"
                ),
                str(message).encode("utf-8"),
                "text/plain",
            )
        )


def get_encoded_body(body: str) -> str:
    try:
        return body.encode("ascii").decode("unicode_escape")
    except UnicodeEncodeError:
        return body


def get_body_from_fragments(body) -> str:
    """Gets a body from the fragments, joined by a double line break"""
    return "\n\n".join(f.content for f in EmailReplyParser.read(body).fragments)


def get_email_body_from_part_payload(part) -> str:
    """Gets an decoded body from the payload part, if the decode fails,
    returns without encoding"""
    try:
        return encoding.smart_str(part.get_payload(decode=True))
    except UnicodeDecodeError:
        return encoding.smart_str(part.get_payload(decode=False))


def attempt_body_extract_from_html(message: str) -> str:
    mail = BeautifulSoup(str(message), "html.parser")
    beautiful_body = mail.find("body")
    body = None
    full_body = None
    if beautiful_body:
        try:
            body = beautiful_body.text
            full_body = body
        except AttributeError:
            pass
    if not body:
        body = ""
    return body, full_body


def mime_content_to_string(
    part: EmailMessage,
) -> str:
    """
    Extract the content from the MIME body part
    :param part: the MIME part to extract the content from
    """
    content_bytes = part.get_payload(decode=True)
    charset = part.get_content_charset()
    # The default for MIME email is 7bit which requires special decoding to utf-8 so make sure
    # we handle the decoding correctly
    if part["Content-Transfer-Encoding"] in [None, "8bit", "7bit"] and (
        charset == "utf-8" or charset is None
    ):
        charset = "unicode_escape"
    content = decodeUnknown(charset, content_bytes)
    return content


def parse_email_content(mime_content: str, is_extract_full_email_msg: bool) -> str:
    if is_extract_full_email_msg:
        # Take the full content including encapsulated "forwarded" and "reply" sections
        return mime_content
    else:
        # Just get the primary part of the email and drop off any text below the actual response text
        return EmailReplyParser.parse_reply(mime_content)


def extract_email_message_content(
    part: MIMEPart,
    files: List,
    include_chained_msgs: bool,
) -> (str, str):
    """
    Uses the get_body() method of the email package to extract the email message content.
    If there is an HTML version of the email message content then it is stored as an attachment.
    If there is a plain text part then that is used for storing the email content against the ticket.
    Otherwise if there is just an HTML part then the HTML is parsed to extract a simple text message.
    There is special handling for the case when a multipart/related part holds the message content when
    there are multiple attachments to the email.
    :param part: the base email MIME part to be searched
    :param files: any MIME parts to be attached are added to this list
    :param include_chained_msgs: flag to indicate if the entire email message content including past
           replies must be extracted
    """
    message_part: MIMEPart = part.get_body()
    # handle the case where there is no content, just an attachment
    if not message_part:
        return None, None
    parent_part: MIMEPart = part
    content_type = message_part.get_content_type()
    # Handle the possibility of a related part formatted email
    if "multipart/related" == content_type:
        # We want the actual message text so try again on the related MIME part
        parent_part = message_part
        message_part = message_part.get_body(
            preferencelist=[
                "html",
                "plain",
            ]
        )
        content_type = message_part.get_content_type()
    mime_content = None
    formatted_body = None  # Retain the original content by using a secondary variable if the HTML needs wrapping
    if "text/html" == content_type:
        # add the HTML message as an attachment wrapping if necessary
        mime_content = mime_content_to_string(message_part)
        if "<body" not in mime_content:
            formatted_body = f"<body>{mime_content}</body>"
        if "<html" not in mime_content:
            formatted_body = f'<html><head><meta charset="utf-8" /></head>\
                               {mime_content if formatted_body is None else formatted_body}</html>'
        files.append(
            SimpleUploadedFile(
                HTML_EMAIL_ATTACHMENT_FILENAME,
                (mime_content if formatted_body is None else formatted_body).encode(
                    "utf-8"
                ),
                "text/html",
            )
        )
        # Try to get a plain part message
        plain_message_part = parent_part.get_body(
            preferencelist=[
                "plain",
            ]
        )
        if plain_message_part:
            # Replace mime_content with the plain text part content
            mime_content = mime_content_to_string(plain_message_part)
            message_part = plain_message_part
            content_type = "text/plain"
        else:
            # Try to constitute the HTML response as plain text
            mime_content, _x = attempt_body_extract_from_html(
                mime_content if formatted_body is None else formatted_body
            )
    else:
        # Is either text/plain or some random content-type so just decode the part content and store as is
        mime_content = mime_content_to_string(message_part)
    # We should now have the mime content
    filtered_body = (
        parse_email_content(mime_content, include_chained_msgs) if mime_content else ""
    )
    if not filtered_body or "" == filtered_body.strip():
        # A unit test that has a different HTML content to plain text which seems an invalid case as email
        # tools should retain the HTML to be consistent with the plain text but manage this as a special case
        # Try to constitute the HTML response as plain text
        if formatted_body:
            filtered_body, _x = attempt_body_extract_from_html(formatted_body)
        else:
            filtered_body = mime_content
    # Only need the full message if the message_body excludes the chained messages
    return filtered_body, mime_content


def process_as_attachment(
    part: MIMEPart, counter: int, files: List, logger: logging.Logger
):
    name = part.get_filename()
    if name:
        name = f"part-{counter}_{email.utils.collapse_rfc2231_value(name)}"
    else:
        ext = mimetypes.guess_extension(part.get_content_type())
        name = f"part-{counter}{ext}"
    # Extract payload accounting for attached multiparts
    payload_bytes = (
        part.as_bytes() if part.is_multipart() else part.get_payload(decode=True)
    )
    files.append(SimpleUploadedFile(name, payload_bytes, mimetypes.guess_type(name)[0]))
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("Processed MIME as attachment: %s", name)
    return


def extract_email_subject(
    email_msg: EmailMessage,
) -> str:
    subject = email_msg.get("subject", _("Comment from e-mail"))
    subject = decode_mail_headers(decodeUnknown(email_msg.get_charset(), subject))
    for affix in STRIPPED_SUBJECT_STRINGS:
        subject = subject.replace(affix, "")
    return subject.strip()


def extract_attachments(
    target_part: MIMEPart,
    files: List,
    logger: logging.Logger,
    counter: int = 1,
    content_parts_excluded: bool = False,
) -> (int, bool):
    """
    If the MIME part is a multipart and not identified as "inline" or "attachment" then
    iterate over the sub parts recursively.
    Otherwise extract MIME part content and add as an attachment.
    It will recursively descend as appropriate ensuring that all parts not part of the message content
    are added to the list of files to be attached. To cater for the possibility of text/plain and text/html parts
    that are further down in the multipart hierarchy than the ones that ones meant to provide that content,
    iterators are selectively used.
    :param part: the email MIME part to be processed
    :param files: any MIME part content or MIME parts to be attached are added to this list
    :param logger: the logger to use for this MIME part processing
    :param counter: the count of MIME parts added as attachment
    :param content_parts_excluded: the MIME part(s) that provided the message content have been excluded
    :returns the count of mime parts added as attachments and a boolean if the content parts have been excluded
    """
    content_type = target_part.get_content_type()
    content_maintype = target_part.get_content_maintype()
    if (
        "multipart" == content_maintype
        and target_part.get_content_disposition() not in ["inline", "attachment"]
    ):
        # Cycle through all MIME parts in the email extracting the attachments that were not part of the message body
        # If this is a "related" multipart then we can use the message part excluder iterator directly
        if "multipart/related" == content_type:
            if content_parts_excluded:
                # This should really never happen in a properly constructed email message but...
                logger.warning(
                    "WARNING! Content type MIME parts have been excluded but a multipart/related has been encountered.\
                     There may be missing information in attachments."
                )
            else:
                content_parts_excluded = True
            # Use the iterator that automatically excludes message content parts
            for part in target_part.iter_attachments():
                counter, content_parts_excluded = extract_attachments(
                    part, files, logger, counter, content_parts_excluded
                )
        # The iterator must be different depending on whether we have already excluded message content parts
        else:
            # Content part might be 1 or 2 parts but will be at same level so track finding at least 1
            content_part_detected = False
            for part in target_part.iter_parts():
                if not content_parts_excluded and part.get_content_type() in [
                    "text/plain",
                    "text/html",
                ]:
                    content_part_detected = True
                    continue
                # Recurse into the part to process embedded parts
                counter, content_parts_excluded = extract_attachments(
                    part, files, logger, counter, content_parts_excluded
                )
            # If we have found 1 or more content parts then flag that the content parts have been omitted
            # to ensure that other text/* parts are attached
            if content_part_detected:
                content_parts_excluded = True
    else:
        process_as_attachment(target_part, counter, files, logger)
        counter = counter + 1
    return (counter, content_parts_excluded)


def extract_email_metadata(
    message: str, queue: Queue, logger: logging.Logger
) -> Ticket:
    """
    Extracts the text/plain  mime part if there is one as the ticket description and
    stores the text/html part as an attachment if it is present.
    If no text/plain  part is present then it will try to use the text/html part if
    it is present as the ticket description by removing the HTML formatting.
    If neither a text/plain or text/html is present then it will use the first text/*
    MIME part that it finds as the ticket description.
    By default it will always take only the actual message and drop any chained messages
    from replies.
    The HELPDESK_FULL_FIRST_MESSAGE_FROM_EMAIL settings can force the entire message to be
    stored in the ticket if it is a new ticket by setting it to True.
    In this scenario, if it is a reply that is a forwarded message with no actual message,
    then the description will be sourced from the text/html part and the forwarded message
    will be in the FollowUp record associated with the ticket.
    It will iterate over every MIME part and store all MIME parts as attachments apart
    from the text/plain part.
    There may be a case for trying to exclude repeated signature images by checking if an
    attachment of the same name already exists as an attachment on the ticket but that is
    not implemented.
    :param message: the raw email message received
    :param queue: the queue that the message is assigned to
    :param logger: the logger to be used
    """
    # 'message' must be an RFC822 formatted message to correctly parse.
    # NBot sure why but policy explicitly set to default is required for any messages with attachments in them
    message_obj: EmailMessage = email.message_from_string(
        message, EmailMessage, policy=policy.default
    )

    subject = extract_email_subject(message_obj)

    sender_email = _("Unknown Sender")
    sender_hdr = message_obj.get("from")
    if sender_hdr:
        # Parse the header which extracts the first email address in the list if more than one
        # The parseaddr method returns a tuple in the form <real name> <email address>
        # Only need the actual email address from the tuple not the "real name"
        # Since the spec requires that all email addresses are ASCII, they will not be encoded
        sender_email = email.utils.parseaddr(sender_hdr)[1]

    for ignore in IgnoreEmail.objects.filter(Q(queues=queue) | Q(queues__isnull=True)):
        if ignore.test(sender_email):
            raise (
                IgnoreTicketException()
                if ignore.keep_in_mailbox
                else DeleteIgnoredTicketException()
            )

    # First try to get ticket ID from the current queue's slug
    ticket_id: typing.Optional[int] = get_ticket_id_from_subject_slug(
        queue.slug, subject, logger
    )

    # If no ticket ID found in the current queue, check all other queues
    original_queue = queue  # Store the original queue in case we need to revert
    if ticket_id is None:
        # Get all enabled queues except the current one, regardless of allow_email_submission
        other_queues = Queue.objects.exclude(id=queue.id).filter(
            email_box_type__isnull=False
        )

        for other_queue in other_queues:
            ticket_id = get_ticket_id_from_subject_slug(
                other_queue.slug, subject, logger
            )
            if ticket_id is not None:
                # Found a matching ticket in another queue, use that queue instead
                logger.info(
                    f"Found ticket {ticket_id} matching subject in queue {other_queue.slug} "
                    f"instead of current queue {queue.slug}"
                )
                queue = other_queue
                break

        # If no ticket ID was found in any queue, revert to the original queue for new ticket
        if ticket_id is None:
            queue = original_queue
            logger.info(
                f"No existing ticket found, reverting to original queue {queue.slug}"
            )

    files = []
    # first message in thread, we save full body to avoid losing forwards and things like that
    include_chained_msgs = (
        True
        if ticket_id is None
        and getattr(settings, "HELPDESK_FULL_FIRST_MESSAGE_FROM_EMAIL", False)
        else False
    )
    filtered_body, full_body = extract_email_message_content(
        message_obj, files, include_chained_msgs
    )
    # If the base part is not a multipart then it will have already been processed as the vbody content so
    # no need to process attachments
    if (
        "multipart" == message_obj.get_content_maintype()
        and helpdesk_settings.HELPDESK_ENABLE_ATTACHMENTS
    ):
        # Find and attach all other parts or part contents as attachments
        counter, content_parts_excluded = extract_attachments(
            message_obj, files, logger
        )
        # Check if there is expected to be a content part
        if not content_parts_excluded and (filtered_body or full_body):
            # Unexpected situation and may mean there is a hole in the email processing logic
            logger.warning(
                "Failed to exclude email content when parsing all MIME parts in the multipart.\
                 Verify that there were no text/* parts containing message content."
            )
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "Email parsed and %s attachments were found and attached.", counter
            )
    if helpdesk_settings.HELPDESK_ENABLE_ATTACHMENTS:
        add_file_if_always_save_incoming_email_message(files, message)

    smtp_priority = message_obj.get("priority", "")
    smtp_importance = message_obj.get("importance", "")
    high_priority_types = {"high", "important", "1", "urgent"}
    priority = 2 if high_priority_types & {smtp_priority, smtp_importance} else 3

    payload = {
        "body": filtered_body,
        "full_body": full_body,
        "subject": subject,
        "queue": queue,
        "sender_email": sender_email,
        "priority": priority,
        "files": files,
    }

    return create_object_from_email_message(
        message_obj, ticket_id, payload, files, logger=logger
    )



from helpdesk.models import CustomField, KBItem, Queue


class AbstractCreateTicketMixin:
    def get_initial(self):
        initial_data = {}
        request = self.request
        try:
            initial_data["queue"] = Queue.objects.get(
                slug=request.GET.get("queue", None)
            ).id
        except Queue.DoesNotExist:
            pass
        u = request.user
        if (
            u.is_authenticated
            and u.usersettings_helpdesk.use_email_as_submitter
            and u.email
        ):
            initial_data["submitter_email"] = u.email

        query_param_fields = ["submitter_email", "title", "body", "queue", "kbitem"]
        custom_fields = [
            "custom_%s" % f.name for f in CustomField.objects.filter(staff_only=False)
        ]
        query_param_fields += custom_fields
        for qpf in query_param_fields:
            initial_data[qpf] = request.GET.get(qpf, initial_data.get(qpf, ""))

        return initial_data

    def get_form_kwargs(self, *args, **kwargs):
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kbitem = self.request.GET.get(
            "kbitem",
            self.request.POST.get("kbitem", None),
        )
        if kbitem:
            try:
                kwargs["kbcategory"] = KBItem.objects.get(pk=int(kbitem)).category
            except (ValueError, KBItem.DoesNotExist):
                pass
        return kwargs



# Generated by Django 2.2.9 on 2020-01-27 16:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0028_kbitem_team"),
    ]

    operations = [
        migrations.AddField(
            model_name="kbcategory",
            name="public",
            field=models.BooleanField(
                default=True, verbose_name="Is KBCategory publicly visible?"
            ),
        ),
    ]



# Generated by Django 2.0.5 on 2019-02-06 13:24

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0023_add_enable_notifications_on_email_events_to_ticket"),
    ]

    operations = [
        migrations.AddField(
            model_name="followup",
            name="time_spent",
            field=models.DurationField(
                blank=True, help_text="Time spent on this follow up", null=True
            ),
        ),
    ]



"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008-2025 Jutda. All Rights Reserved. See LICENSE for details.

views/public.py - All public facing views, eg non-staff (no authentication
                  required) views.
"""

from django.conf import settings
from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    PermissionDenied,
)
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.decorators.clickjacking import xframe_options_exempt
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import TemplateView
from django.views.generic.edit import FormView
from helpdesk import settings as helpdesk_settings
from helpdesk.decorators import is_helpdesk_staff, protect_view
from helpdesk.lib import text_is_spam
from helpdesk.models import Queue, Ticket, UserSettings
from helpdesk.user import huser_from_request
import helpdesk.views.abstract_views as abstract_views
import helpdesk.views.staff as staff
from importlib import import_module
import logging
from urllib.parse import quote


logger = logging.getLogger(__name__)


def create_ticket(request, *args, **kwargs):
    if is_helpdesk_staff(request.user):
        return staff.CreateTicketView.as_view()(request, *args, **kwargs)
    else:
        return protect_view(CreateTicketView.as_view())(request, *args, **kwargs)


class BaseCreateTicketView(abstract_views.AbstractCreateTicketMixin, FormView):
    def get_form_class(self):
        try:
            the_module, the_form_class = (
                helpdesk_settings.HELPDESK_PUBLIC_TICKET_FORM_CLASS.rsplit(".", 1)
            )
            the_module = import_module(the_module)
            the_form_class = getattr(the_module, the_form_class)
        except Exception as e:
            raise ImproperlyConfigured(
                f"Invalid custom form class {helpdesk_settings.HELPDESK_PUBLIC_TICKET_FORM_CLASS}"
            ) from e
        return the_form_class

    def dispatch(self, *args, **kwargs):
        request = self.request
        if (
            not request.user.is_authenticated
            and helpdesk_settings.HELPDESK_REDIRECT_TO_LOGIN_BY_DEFAULT
        ):
            return HttpResponseRedirect(reverse("login"))

        if is_helpdesk_staff(request.user) or (
            request.user.is_authenticated
            and helpdesk_settings.HELPDESK_ALLOW_NON_STAFF_TICKET_UPDATE
        ):
            try:
                if request.user.usersettings_helpdesk.login_view_ticketlist:
                    return HttpResponseRedirect(reverse("helpdesk:list"))
                else:
                    return HttpResponseRedirect(reverse("helpdesk:dashboard"))
            except UserSettings.DoesNotExist:
                return HttpResponseRedirect(reverse("helpdesk:dashboard"))
        return super().dispatch(*args, **kwargs)

    def get_initial(self):
        initial_data = super().get_initial()

        # add pre-defined data for public ticket
        if hasattr(settings, "HELPDESK_PUBLIC_TICKET_QUEUE"):
            # get the requested queue; return an error if queue not found
            try:
                initial_data["queue"] = Queue.objects.get(
                    slug=settings.HELPDESK_PUBLIC_TICKET_QUEUE,
                    allow_public_submission=True,
                ).id
            except Queue.DoesNotExist as e:
                logger.fatal(
                    "Public queue '%s' is configured as default but can't be found",
                    settings.HELPDESK_PUBLIC_TICKET_QUEUE,
                )
                raise ImproperlyConfigured("Wrong public queue configuration") from e
        if hasattr(settings, "HELPDESK_PUBLIC_TICKET_PRIORITY"):
            initial_data["priority"] = settings.HELPDESK_PUBLIC_TICKET_PRIORITY
        if hasattr(settings, "HELPDESK_PUBLIC_TICKET_DUE_DATE"):
            initial_data["due_date"] = settings.HELPDESK_PUBLIC_TICKET_DUE_DATE
        return initial_data

    def get_form_kwargs(self, *args, **kwargs):
        kwargs = super().get_form_kwargs(*args, **kwargs)
        if "_hide_fields_" in self.request.GET:
            kwargs["hidden_fields"] = self.request.GET.get("_hide_fields_", "").split(
                ","
            )
        kwargs["readonly_fields"] = self.request.GET.get("_readonly_fields_", "").split(
            ","
        )
        return kwargs

    def form_valid(self, form):
        request = self.request
        if text_is_spam(form.cleaned_data["body"], request):
            # This submission is spam. Let's not save it.
            return render(request, template_name="helpdesk/public_spam.html")
        else:
            ticket = form.save(
                user=self.request.user if self.request.user.is_authenticated else None
            )
            try:
                return HttpResponseRedirect(
                    "%s?ticket=%s&email=%s&key=%s"
                    % (
                        reverse("helpdesk:public_view"),
                        ticket.ticket_for_url,
                        quote(ticket.submitter_email),
                        ticket.secret_key,
                    )
                )
            except ValueError:
                # if someone enters a non-int string for the ticket
                return HttpResponseRedirect(reverse("helpdesk:home"))


class CreateTicketIframeView(BaseCreateTicketView):
    template_name = "helpdesk/public_create_ticket_iframe.html"

    @csrf_exempt
    @xframe_options_exempt
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def form_valid(self, form):
        if super().form_valid(form).status_code == 302:
            return HttpResponseRedirect(reverse("helpdesk:success_iframe"))


class SuccessIframeView(TemplateView):
    template_name = "helpdesk/success_iframe.html"

    @xframe_options_exempt
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)


class CreateTicketView(BaseCreateTicketView):
    template_name = "helpdesk/public_create_ticket.html"

    def get_form(self, form_class=None):
        form = super().get_form(form_class)
        # Add the CSS error class to the form in order to better see them in
        # the page
        form.error_css_class = "text-danger"
        return form


class Homepage(CreateTicketView):
    template_name = "helpdesk/public_homepage.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["kb_categories"] = huser_from_request(
            self.request
        ).get_allowed_kb_categories()
        return context


class SearchForTicketView(TemplateView):
    template_name = "helpdesk/public_view_form.html"

    def get(self, request, *args, **kwargs):
        if (
            hasattr(settings, "HELPDESK_VIEW_A_TICKET_PUBLIC")
            and settings.HELPDESK_VIEW_A_TICKET_PUBLIC
        ):
            context = self.get_context_data(**kwargs)
            return self.render_to_response(context)
        else:
            raise PermissionDenied(
                "Public viewing of tickets without a secret key is forbidden."
            )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        request = self.request
        email = request.GET.get("email", None)
        error_message = kwargs.get("error_message", None)

        context.update(
            {
                "ticket": False,
                "email": email,
                "error_message": error_message,
                "helpdesk_settings": helpdesk_settings,
            }
        )
        return context


class ViewTicket(TemplateView):
    template_name = "helpdesk/public_view_ticket.html"

    def get(self, request, *args, **kwargs):
        ticket_req = request.GET.get("ticket", None)
        email = request.GET.get("email", None)
        key = request.GET.get("key", "")

        if not (ticket_req and email):
            if ticket_req is None and email is None:
                return SearchForTicketView.as_view()(request)
            else:
                return SearchForTicketView.as_view()(
                    request, _("Missing ticket ID or e-mail address. Please try again.")
                )

        try:
            queue, ticket_id = Ticket.queue_and_id_from_query(ticket_req)
            if request.user.is_authenticated and request.user.email == email:
                ticket = Ticket.objects.get(id=ticket_id, submitter_email__iexact=email)
            elif (
                hasattr(settings, "HELPDESK_VIEW_A_TICKET_PUBLIC")
                and settings.HELPDESK_VIEW_A_TICKET_PUBLIC
            ):
                ticket = Ticket.objects.get(id=ticket_id, submitter_email__iexact=email)
            else:
                ticket = Ticket.objects.get(
                    id=ticket_id, submitter_email__iexact=email, secret_key__iexact=key
                )
        except (ObjectDoesNotExist, ValueError):
            return SearchForTicketView.as_view()(
                request, _("Invalid ticket ID or e-mail address. Please try again.")
            )

        if "close" in request.GET and ticket.status == Ticket.RESOLVED_STATUS:
            from helpdesk.update_ticket import update_ticket

            update_ticket(
                request.user,
                ticket,
                public=True,
                comment=_("Submitter accepted resolution and closed ticket"),
                new_status=Ticket.CLOSED_STATUS,
            )
            return HttpResponseRedirect(ticket.ticket_url)

        # Prepare context for rendering
        context = {
            "key": key,
            "mail": email,
            "ticket": ticket,
            "helpdesk_settings": helpdesk_settings,
            "next": self.get_next_url(ticket_id),
        }
        return self.render_to_response(context)

    def get_next_url(self, ticket_id):
        redirect_url = ""
        if is_helpdesk_staff(self.request.user):
            redirect_url = reverse("helpdesk:view", args=[ticket_id])
            if "close" in self.request.GET:
                redirect_url += "?close"
        elif helpdesk_settings.HELPDESK_NAVIGATION_ENABLED:
            redirect_url = reverse("helpdesk:view", args=[ticket_id])
        return redirect_url


class MyTickets(TemplateView):
    template_name = "helpdesk/my_tickets.html"

    def get(self, request, *args, **kwargs):
        if not request.user.is_authenticated:
            return HttpResponseRedirect(reverse("helpdesk:login"))

        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)


def change_language(request):
    return_to = ""
    if "return_to" in request.GET:
        return_to = request.GET["return_to"]

    return render(request, "helpdesk/public_change_language.html", {"next": return_to})



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-03-01 19:43
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0022_add_submitter_email_id_field_to_ticket"),
    ]

    operations = [
        migrations.AddField(
            model_name="queue",
            name="enable_notifications_on_email_events",
            field=models.BooleanField(
                default=False,
                help_text="When an email arrives to either create a ticket or to interact with an existing discussion. Should email notifications be sent ? Note: the new_ticket_cc and updated_ticket_cc work independently of this feature",
                verbose_name="Notify contacts when email updates arrive",
            ),
        ),
    ]



""" """

from django.db import models


class BaseModel(models.Model):
    """
    Allows adding common behaviour to the classes across the app.
    A common usage would be to define a custom id field like this:
        id = models.UUIDField(
            primary_key=True,
            editable=False,
        )


    """

    class Meta:
        abstract = True



"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

templatetags/load_helpdesk_settings.py - returns the settings as defined in
                                    django-helpdesk/helpdesk/settings.py
"""

from django.template import Library
from helpdesk import settings as helpdesk_settings_config


def load_helpdesk_settings(request):
    try:
        return helpdesk_settings_config
    except Exception as e:
        import sys

        print(
            "'load_helpdesk_settings' template tag (django-helpdesk) crashed with following error:",
            file=sys.stderr,
        )
        print(e, file=sys.stderr)
        return ""


register = Library()
register.filter("load_helpdesk_settings", load_helpdesk_settings)



from base64 import b64decode, b64encode
from django.db.models import Q, Max
from django.db.models import F, Window, Subquery, OuterRef
from .models import FollowUp
from django.urls import reverse
from django.utils.html import escape
from django.utils.translation import gettext as _
from helpdesk.serializers import DatatablesTicketSerializer
import json
from model_utils import Choices


def query_to_base64(query):
    """
    Converts a query dict object to a base64-encoded bytes object.
    """
    return b64encode(json.dumps(query).encode("UTF-8")).decode("ascii")


def query_from_base64(b64data):
    """
    Converts base64-encoded bytes object back to a query dict object.
    """
    query = {"search_string": ""}
    query.update(json.loads(b64decode(b64data).decode("utf-8")))
    if query["search_string"] is None:
        query["search_string"] = ""
    return query


def get_search_filter_args(search):
    if not search:
        return Q()
    if search.startswith("queue:"):
        return Q(queue__title__icontains=search[len("queue:") :])
    if search.startswith("priority:"):
        return Q(priority__icontains=search[len("priority:") :])
    my_filter = Q()
    for subsearch in search.split("OR"):
        subsearch = subsearch.strip()
        if not subsearch:
            continue
        my_filter |= (
            Q(id__icontains=subsearch)
            | Q(title__icontains=subsearch)
            | Q(description__icontains=subsearch)
            | Q(priority__icontains=subsearch)
            | Q(resolution__icontains=subsearch)
            | Q(submitter_email__icontains=subsearch)
            | Q(assigned_to__email__icontains=subsearch)
            | Q(ticketcustomfieldvalue__value__icontains=subsearch)
            | Q(created__icontains=subsearch)
            | Q(due_date__icontains=subsearch)
        )
    return my_filter


DATATABLES_ORDER_COLUMN_CHOICES = Choices(
    ("0", "id"),
    ("1", "title"),
    ("2", "priority"),
    ("3", "queue"),
    ("4", "status"),
    ("5", "created"),
    ("6", "due_date"),
    ("7", "assigned_to"),
    ("8", "submitter_email"),
    ("9", "last_followup"),
    # ('10', 'time_spent'),
    ("11", "kbitem"),
)


DATATABLES_COLUMN_NUM_LOOKUP = {v: k for k, v in DATATABLES_ORDER_COLUMN_CHOICES}


def get_query_class():
    from django.conf import settings

    def _get_query_class():
        return __Query__

    return getattr(settings, "HELPDESK_QUERY_CLASS", _get_query_class)()


class __Query__:
    def __init__(self, huser, base64query=None, query_params=None):
        self.huser = huser
        self.params = query_params if query_params else query_from_base64(base64query)
        self.base64 = base64query if base64query else query_to_base64(query_params)
        self.result = None

    def get_search_filter_args(self):
        search = self.params.get("search_string", "")
        return get_search_filter_args(search)

    def __run__(self, queryset):
        """
        Apply a dict-based set of value_filters & parameters to a queryset.

        queryset is a Django queryset, eg MyModel.objects.all() or
            MyModel.objects.filter(user=request.user)

        params is a dictionary that contains the following:
           filtering: A dict of Django ORM value_filters, eg:
            {'user__id__in': [1, 3, 103], 'title__contains': 'foo'}

        search_string: A freetext search string

        sorting: The name of the column to sort by
        """
        q_args = []
        value_filters = self.params.get("filtering", {})
        null_filters = self.params.get("filtering_null", {})
        if null_filters:
            if value_filters:
                # Check if any of the value value_filters are for the same field as the
                # ISNULL filter so that an OR filter can be set up
                matched_null_keys = []
                for null_key in null_filters:
                    field_path = null_key[:-8]  # Chop off the "__isnull"
                    matched_key = None
                    for val_key in value_filters:
                        if val_key.startswith(field_path):
                            matched_key = val_key
                            break
                    if matched_key:
                        # Remove the matching filters into a Q param
                        matched_null_keys.append(null_key)
                        # Create an OR query for the selected value(s) OR if the field is NULL
                        v = {}
                        v[val_key] = value_filters[val_key]
                        n = {}
                        n[null_key] = null_filters[null_key]
                        q_args.append((Q(**v) | Q(**n)))
                        del value_filters[matched_key]
                # Now remove the matched null keys
                for null_key in matched_null_keys:
                    del null_filters[null_key]
        queryset = queryset.filter(
            *q_args,
            (Q(**value_filters) & Q(**null_filters)) & self.get_search_filter_args(),
        )
        sorting = self.params.get("sorting", None)
        if sorting:
            sortreverse = self.params.get("sortreverse", None)
            if sortreverse:
                sorting = "-%s" % sorting
            queryset = queryset.order_by(sorting)
        # https://stackoverflow.com/questions/30487056/django-queryset-contains-duplicate-entries
        return queryset.distinct()

    def get(self):
        # Prefilter the allowed tickets
        tickets = self.huser.get_tickets_in_queues().select_related()
        return self.__run__(tickets)

    def get_datatables_context(self, *, column_lookup=None, **kwargs):
        """
        This function takes in a list of ticket objects from the views and throws it
        to the datatables on ticket_list.html. If a search string was entered, this
        function filters existing dataset on search string and returns a filtered
        filtered list. The `draw`, `length` etc parameters are for datatables to
        display meta data on the table contents. The returning queryset is passed
        to a Serializer called DatatablesTicketSerializer in serializers.py.
        Optionally, one can pass a dictionary in to override the default column
        mapping for sorting
        """
        objects = self.get()
        draw = int(kwargs.get("draw", [0])[0])
        length = int(kwargs.get("length", [25])[0])
        start = int(kwargs.get("start", [0])[0])
        search_value = kwargs.get("search[value]", [""])[0]
        if column_lookup is None:
            column_lookup = DATATABLES_ORDER_COLUMN_CHOICES
            num_lookup = DATATABLES_COLUMN_NUM_LOOKUP
        else:
            num_lookup = {v: k for k, v in column_lookup}

        sorting = self.params.get("sorting", "created")
        default_order_col = num_lookup.get(sorting, "5")
        sortreverse = self.params.get("sortreverse", None)
        default_order = "desc" if sortreverse else "asc"

        order_column = kwargs.get("order[0][column]", [default_order_col])[0]
        order = kwargs.get("order[0][dir]", [default_order])[0]

        order_column = column_lookup[order_column]
        # django orm '-' -> desc
        if order == "desc":
            order_column = "-" + order_column

        queryset = objects.annotate(
            last_followup=Subquery(
                FollowUp.objects.order_by()
                .annotate(
                    last_followup=Window(
                        expression=Max("date"),
                        partition_by=[
                            F("ticket_id"),
                        ],
                        order_by="-date",
                    )
                )
                .filter(ticket_id=OuterRef("id"))
                .values("last_followup")
                .distinct()
            )
        )

        total = queryset.count()

        if search_value:  # Dead code currently
            queryset = queryset.filter(get_search_filter_args(search_value))

        count = queryset.count()
        queryset = queryset.order_by(order_column)[start : start + length]
        return {
            "data": DatatablesTicketSerializer(queryset, many=True).data,
            "recordsFiltered": count,
            "recordsTotal": total,
            "draw": draw,
        }

    def get_timeline_context(self):
        events = []

        for ticket in self.get():
            for followup in ticket.followup_set.all():
                event = {
                    "start_date": self.mk_timeline_date(followup.date),
                    "text": {
                        "headline": ticket.title + " - " + followup.title,
                        "text": (
                            (
                                escape(followup.comment)
                                if followup.comment
                                else _("No text")
                            )
                            + '<br/> <a href="%s" class="btn" role="button">%s</a>'
                            % (
                                reverse(
                                    "helpdesk:view", kwargs={"ticket_id": ticket.pk}
                                ),
                                _("View ticket"),
                            )
                        ),
                    },
                    "group": _("Messages"),
                }
                events.append(event)

        return {
            "events": events,
        }

    def mk_timeline_date(self, date):
        return {
            "year": date.year,
            "month": date.month,
            "day": date.day,
            "hour": date.hour,
            "minute": date.minute,
            "second": date.second,
        }



# Generated by Django 2.2.6 on 2020-01-07 12:21

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("helpdesk", "0026_kbitem_attachments"),
    ]

    operations = [
        migrations.AddField(
            model_name="kbcategory",
            name="queue",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="helpdesk.Queue",
                verbose_name="Default queue when creating a ticket after viewing this category.",
            ),
        ),
        migrations.AddField(
            model_name="kbitem",
            name="downvoted_by",
            field=models.ManyToManyField(
                related_name="downvotes", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="ticket",
            name="kbitem",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="helpdesk.KBItem",
                verbose_name="Knowledge base item the user was viewing when they created this ticket.",
            ),
        ),
        migrations.AlterField(
            model_name="followupattachment",
            name="filename",
            field=models.CharField(
                blank=True, max_length=1000, verbose_name="Filename"
            ),
        ),
        migrations.AlterField(
            model_name="followupattachment",
            name="mime_type",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="MIME Type"
            ),
        ),
        migrations.AlterField(
            model_name="followupattachment",
            name="size",
            field=models.IntegerField(
                blank=True, help_text="Size of this file in bytes", verbose_name="Size"
            ),
        ),
        migrations.AlterField(
            model_name="kbiattachment",
            name="filename",
            field=models.CharField(
                blank=True, max_length=1000, verbose_name="Filename"
            ),
        ),
        migrations.AlterField(
            model_name="kbiattachment",
            name="mime_type",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="MIME Type"
            ),
        ),
        migrations.AlterField(
            model_name="kbiattachment",
            name="size",
            field=models.IntegerField(
                blank=True, help_text="Size of this file in bytes", verbose_name="Size"
            ),
        ),
        migrations.AlterField(
            model_name="kbitem",
            name="voted_by",
            field=models.ManyToManyField(
                related_name="votes", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AlterField(
            model_name="queue",
            name="enable_notifications_on_email_events",
            field=models.BooleanField(
                blank=True,
                default=False,
                help_text="When an email arrives to either create a ticket or to interact with an existing discussion. Should email notifications be sent ? Note: the new_ticket_cc and updated_ticket_cc work independently of this feature",
                verbose_name="Notify contacts when email updates arrive",
            ),
        ),
    ]



# Generated by Django 3.2.16 on 2023-03-25 15:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0036_add_attachment_validator"),
    ]

    operations = [
        migrations.AlterField(
            model_name="queue",
            name="email_box_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("pop3", "POP 3"),
                    ("imap", "IMAP"),
                    ("oauth", "IMAP OAUTH"),
                    ("local", "Local Directory"),
                ],
                help_text="E-Mail server type for creating tickets automatically from a mailbox - both POP3 and IMAP are supported, as well as reading from a local directory.",
                max_length=5,
                null=True,
                verbose_name="E-Mail Box Type",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0004_add_per_queue_staff_membership"),
    ]

    operations = [
        migrations.AlterField(
            model_name="escalationexclusion",
            name="queues",
            field=models.ManyToManyField(
                help_text="Leave blank for this exclusion to be applied to all queues, or select those queues you wish to exclude with this entry.",
                to="helpdesk.Queue",
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="ignoreemail",
            name="queues",
            field=models.ManyToManyField(
                help_text="Leave blank for this e-mail to be ignored on all queues, or select those queues you wish to ignore this e-mail for.",
                to="helpdesk.Queue",
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="presetreply",
            name="queues",
            field=models.ManyToManyField(
                help_text="Leave blank to allow this reply to be used for all queues, or select those queues you wish to limit this reply to.",
                to="helpdesk.Queue",
                blank=True,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0010_remove_queuemembership"),
    ]

    operations = [
        migrations.AlterField(
            model_name="queue",
            name="permission_name",
            field=models.CharField(
                editable=False,
                max_length=50,
                blank=True,
                help_text="Name used in the django.contrib.auth permission system",
                null=True,
                verbose_name="Django auth permission name",
            ),
        ),
        migrations.AlterField(
            model_name="queue",
            name="slug",
            field=models.SlugField(
                help_text="This slug is used when building ticket ID's. Once set, try not to change it or e-mailing may get messy.",
                unique=True,
                verbose_name="Slug",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0006_email_maxlength"),
    ]

    operations = [
        migrations.AlterField(
            model_name="customfield",
            name="label",
            field=models.CharField(
                help_text="The display label for this field",
                max_length=30,
                verbose_name="Label",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2017-03-08 17:51
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0015_expand_permission_name_size"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="attachment",
            options={
                "ordering": ("filename",),
                "verbose_name": "Attachment",
                "verbose_name_plural": "Attachments",
            },
        ),
        migrations.AlterModelOptions(
            name="emailtemplate",
            options={
                "ordering": ("template_name", "locale"),
                "verbose_name": "e-mail template",
                "verbose_name_plural": "e-mail templates",
            },
        ),
        migrations.AlterModelOptions(
            name="followup",
            options={
                "ordering": ("date",),
                "verbose_name": "Follow-up",
                "verbose_name_plural": "Follow-ups",
            },
        ),
        migrations.AlterModelOptions(
            name="kbcategory",
            options={
                "ordering": ("title",),
                "verbose_name": "Knowledge base category",
                "verbose_name_plural": "Knowledge base categories",
            },
        ),
        migrations.AlterModelOptions(
            name="kbitem",
            options={
                "ordering": ("title",),
                "verbose_name": "Knowledge base item",
                "verbose_name_plural": "Knowledge base items",
            },
        ),
        migrations.AlterModelOptions(
            name="presetreply",
            options={
                "ordering": ("name",),
                "verbose_name": "Pre-set reply",
                "verbose_name_plural": "Pre-set replies",
            },
        ),
        migrations.AlterUniqueTogether(
            name="ticketcustomfieldvalue",
            unique_together=set([("ticket", "field")]),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0013_email_box_local_dir_and_logging"),
    ]

    operations = [
        migrations.AlterField(
            model_name="usersettings",
            name="user",
            field=models.OneToOneField(
                to=settings.AUTH_USER_MODEL,
                related_name="usersettings_helpdesk",
                on_delete=models.CASCADE,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-10 19:27
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0014_usersettings_related_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="queue",
            name="permission_name",
            field=models.CharField(
                blank=True,
                editable=False,
                help_text="Name used in the django.contrib.auth permission system",
                max_length=72,
                null=True,
                verbose_name="Django auth permission name",
            ),
        ),
    ]



"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

templatetags/saved_queries.py - This template tag returns previously saved
                                queries. Therefore you don't need to modify
                                any views.
"""

from django import template
from django.db.models import Q
from helpdesk.models import SavedSearch


register = template.Library()


@register.filter
def saved_queries(user):
    try:
        filters = Q(shared__exact=True)
        if user.is_authenticated:
            filters |= Q(user=user)
        user_saved_queries = SavedSearch.objects.filter(filters)
        return user_saved_queries
    except Exception as e:
        import sys

        print(
            "'saved_queries' template tag (django-helpdesk) crashed with following error:",
            file=sys.stderr,
        )
        print(e, file=sys.stderr)
        return ""



"""
Default settings for django-helpdesk.

"""

from django import forms
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _
import os
import re
import warnings
import sys


DEFAULT_USER_SETTINGS = {
    "login_view_ticketlist": True,
    "email_on_ticket_change": True,
    "email_on_ticket_assign": True,
    "tickets_per_page": 25,
    "use_email_as_submitter": True,
}

try:
    DEFAULT_USER_SETTINGS.update(settings.HELPDESK_DEFAULT_SETTINGS)
except AttributeError:
    pass


HAS_TAG_SUPPORT = False

# Use international timezones
USE_TZ: bool = True

# check for secure cookie support
if os.environ.get("SECURE_PROXY_SSL_HEADER"):
    SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")

    SESSION_COOKIE_SECURE = True
    CSRF_COOKIE_SECURE = True

##########################################
# generic options - visible on all pages #
##########################################

# redirect to login page instead of the default homepage when users visits "/"?
HELPDESK_REDIRECT_TO_LOGIN_BY_DEFAULT = getattr(
    settings, "HELPDESK_REDIRECT_TO_LOGIN_BY_DEFAULT", False
)

HELPDESK_PUBLIC_VIEW_PROTECTOR = getattr(
    settings,
    "HELPDESK_PUBLIC_VIEW_PROTECTOR",
    lambda _: None,  # noqa
)

HELPDESK_STAFF_VIEW_PROTECTOR = getattr(
    settings,
    "HELPDESK_STAFF_VIEW_PROTECTOR",
    lambda _: None,  # noqa
)

# Enable ticket and Email attachments
#
# Caution! Set this to False, unless you have secured access to
#   the uploaded files. Otherwise anyone on the Internet will be
#   able to download your ticket attachments.
HELPDESK_ENABLE_ATTACHMENTS = getattr(settings, "HELPDESK_ENABLE_ATTACHMENTS", True)

# Enable the Dependencies field on ticket view
HELPDESK_ENABLE_DEPENDENCIES_ON_TICKET = getattr(
    settings, "HELPDESK_ENABLE_DEPENDENCIES_ON_TICKET", True
)

# Enable the Time spent on field on ticket view
HELPDESK_ENABLE_TIME_SPENT_ON_TICKET = getattr(
    settings, "HELPDESK_ENABLE_TIME_SPENT_ON_TICKET", True
)

# raises a 404 to anon users. It's like it was invisible
HELPDESK_ANON_ACCESS_RAISES_404 = getattr(
    settings, "HELPDESK_ANON_ACCESS_RAISES_404", False
)

# Disable Timeline on ticket list
HELPDESK_TICKETS_TIMELINE_ENABLED = getattr(
    settings, "HELPDESK_TICKETS_TIMELINE_ENABLED", True
)

# show extended navigation by default, to all users, irrespective of staff
# status?
HELPDESK_NAVIGATION_ENABLED = getattr(settings, "HELPDESK_NAVIGATION_ENABLED", False)

# Show the "My Tickets" navigation option for staff members - typically this is for when
# staff can create tickets to action other staff.
HELPDESK_SHOW_MY_TICKETS_IN_NAV_FOR_STAFF = getattr(
    settings, "HELPDESK_SHOW_MY_TICKETS_IN_NAV_FOR_STAFF", True
)

# use public CDNs to serve jquery and other javascript by default?
# otherwise, use built-in static copy
HELPDESK_USE_CDN = getattr(settings, "HELPDESK_USE_CDN", False)

# show dropdown list of languages that ticket comments can be translated into?
HELPDESK_TRANSLATE_TICKET_COMMENTS = getattr(
    settings, "HELPDESK_TRANSLATE_TICKET_COMMENTS", False
)

# list of languages to offer. if set to false,
# all default google translate languages will be shown.
HELPDESK_TRANSLATE_TICKET_COMMENTS_LANG = getattr(
    settings,
    "HELPDESK_TRANSLATE_TICKET_COMMENTS_LANG",
    ["en", "de", "es", "fr", "it", "ru"],
)

# show link to 'change password' on 'User Settings' page?
HELPDESK_SHOW_CHANGE_PASSWORD = getattr(
    settings, "HELPDESK_SHOW_CHANGE_PASSWORD", False
)

# allow user to override default layout for 'followups' - work in progress.
HELPDESK_FOLLOWUP_MOD = getattr(settings, "HELPDESK_FOLLOWUP_MOD", False)

# auto-subscribe user to ticket if (s)he responds to a ticket?
HELPDESK_AUTO_SUBSCRIBE_ON_TICKET_RESPONSE = getattr(
    settings, "HELPDESK_AUTO_SUBSCRIBE_ON_TICKET_RESPONSE", False
)

# URL schemes that are allowed within links
ALLOWED_URL_SCHEMES = getattr(
    settings,
    "ALLOWED_URL_SCHEMES",
    (
        "file",
        "ftp",
        "ftps",
        "http",
        "https",
        "irc",
        "mailto",
        "sftp",
        "ssh",
        "tel",
        "telnet",
        "tftp",
        "vnc",
        "xmpp",
    ),
)

# Ticket status choices
OPEN_STATUS = getattr(settings, "HELPDESK_TICKET_OPEN_STATUS", 1)
REOPENED_STATUS = getattr(settings, "HELPDESK_TICKET_REOPENED_STATUS", 2)
RESOLVED_STATUS = getattr(settings, "HELPDESK_TICKET_RESOLVED_STATUS", 3)
CLOSED_STATUS = getattr(settings, "HELPDESK_TICKET_CLOSED_STATUS", 4)
DUPLICATE_STATUS = getattr(settings, "HELPDESK_TICKET_DUPLICATE_STATUS", 5)

DEFAULT_TICKET_STATUS_CHOICES = (
    (OPEN_STATUS, _("Open")),
    (REOPENED_STATUS, _("Reopened")),
    (RESOLVED_STATUS, _("Resolved")),
    (CLOSED_STATUS, _("Closed")),
    (DUPLICATE_STATUS, _("Duplicate")),
)
TICKET_STATUS_CHOICES = getattr(
    settings, "HELPDESK_TICKET_STATUS_CHOICES", DEFAULT_TICKET_STATUS_CHOICES
)

# List of status choices considered as "open"
DEFAULT_TICKET_OPEN_STATUSES = (OPEN_STATUS, REOPENED_STATUS)
TICKET_OPEN_STATUSES = getattr(
    settings, "HELPDESK_TICKET_OPEN_STATUSES", DEFAULT_TICKET_OPEN_STATUSES
)

# New status list choices depending on current ticket status
DEFAULT_TICKET_STATUS_CHOICES_FLOW = {
    OPEN_STATUS: (
        OPEN_STATUS,
        RESOLVED_STATUS,
        CLOSED_STATUS,
        DUPLICATE_STATUS,
    ),
    REOPENED_STATUS: (
        REOPENED_STATUS,
        RESOLVED_STATUS,
        CLOSED_STATUS,
        DUPLICATE_STATUS,
    ),
    RESOLVED_STATUS: (
        REOPENED_STATUS,
        RESOLVED_STATUS,
        CLOSED_STATUS,
    ),
    CLOSED_STATUS: (
        REOPENED_STATUS,
        CLOSED_STATUS,
    ),
    DUPLICATE_STATUS: (
        REOPENED_STATUS,
        DUPLICATE_STATUS,
    ),
}
TICKET_STATUS_CHOICES_FLOW = getattr(
    settings, "HELPDESK_TICKET_STATUS_CHOICES_FLOW", DEFAULT_TICKET_STATUS_CHOICES_FLOW
)

# Ticket priority choices
DEFAULT_TICKET_PRIORITY_CHOICES = (
    (1, _("1. Critical")),
    (2, _("2. High")),
    (3, _("3. Normal")),
    (4, _("4. Low")),
    (5, _("5. Very Low")),
)
TICKET_PRIORITY_CHOICES = getattr(
    settings, "HELPDESK_TICKET_PRIORITY_CHOICES", DEFAULT_TICKET_PRIORITY_CHOICES
)


#########################
# time tracking options #
#########################

# Follow-ups automatic time_spent calculation
FOLLOWUP_TIME_SPENT_AUTO = getattr(settings, "HELPDESK_FOLLOWUP_TIME_SPENT_AUTO", False)

# Calculate time_spent according to open hours
FOLLOWUP_TIME_SPENT_OPENING_HOURS = getattr(
    settings, "HELPDESK_FOLLOWUP_TIME_SPENT_OPENING_HOURS", {}
)

# Holidays don't count for time_spent calculation
FOLLOWUP_TIME_SPENT_EXCLUDE_HOLIDAYS = getattr(
    settings, "HELPDESK_FOLLOWUP_TIME_SPENT_EXCLUDE_HOLIDAYS", ()
)

# Time doesn't count for listed ticket statuses
FOLLOWUP_TIME_SPENT_EXCLUDE_STATUSES = getattr(
    settings, "HELPDESK_FOLLOWUP_TIME_SPENT_EXCLUDE_STATUSES", ()
)

# Time doesn't count for listed queues slugs
FOLLOWUP_TIME_SPENT_EXCLUDE_QUEUES = getattr(
    settings, "HELPDESK_FOLLOWUP_TIME_SPENT_EXCLUDE_QUEUES", ()
)

############################
# options for public pages #
############################

# show 'view a ticket' section on public page?
HELPDESK_VIEW_A_TICKET_PUBLIC = getattr(settings, "HELPDESK_VIEW_A_TICKET_PUBLIC", True)

# show 'submit a ticket' section on public page?
HELPDESK_SUBMIT_A_TICKET_PUBLIC = getattr(
    settings, "HELPDESK_SUBMIT_A_TICKET_PUBLIC", True
)

# change that to custom class to have extra fields or validation (like captcha)
HELPDESK_PUBLIC_TICKET_FORM_CLASS = getattr(
    settings, "HELPDESK_PUBLIC_TICKET_FORM_CLASS", "helpdesk.forms.PublicTicketForm"
)

# Custom fields constants
CUSTOMFIELD_TO_FIELD_DICT = {
    "boolean": forms.BooleanField,
    "date": forms.DateField,
    "time": forms.TimeField,
    "datetime": forms.DateTimeField,
    "email": forms.EmailField,
    "url": forms.URLField,
    "ipaddress": forms.GenericIPAddressField,
    "slug": forms.SlugField,
}
CUSTOMFIELD_DATE_FORMAT = "%Y-%m-%d"
CUSTOMFIELD_TIME_FORMAT = "%H:%M:%S"
CUSTOMFIELD_DATETIME_FORMAT = f"{CUSTOMFIELD_DATE_FORMAT}T%H:%M"


###################################
# options for update_ticket views #
###################################

""" options for update_ticket views """
# allow non-staff users to interact with tickets?
# can be True/False or a callable accepting the active user and returning
# True if they must be considered helpdesk staff
HELPDESK_ALLOW_NON_STAFF_TICKET_UPDATE = getattr(
    settings, "HELPDESK_ALLOW_NON_STAFF_TICKET_UPDATE", False
)
if not (
    HELPDESK_ALLOW_NON_STAFF_TICKET_UPDATE in (True, False)
    or callable(HELPDESK_ALLOW_NON_STAFF_TICKET_UPDATE)
):
    warnings.warn(
        "HELPDESK_ALLOW_NON_STAFF_TICKET_UPDATE should be set to either True/False or a callable.",
        RuntimeWarning,
    )

# show edit buttons in ticket follow ups.
HELPDESK_SHOW_EDIT_BUTTON_FOLLOW_UP = getattr(
    settings, "HELPDESK_SHOW_EDIT_BUTTON_FOLLOW_UP", True
)

HELPDESK_SHOW_CUSTOM_FIELDS_FOLLOW_UP_LIST = getattr(
    settings, "HELPDESK_SHOW_CUSTOM_FIELDS_FOLLOW_UP_LIST", []
)

# show delete buttons in ticket follow ups if user is 'superuser'
HELPDESK_SHOW_DELETE_BUTTON_SUPERUSER_FOLLOW_UP = getattr(
    settings, "HELPDESK_SHOW_DELETE_BUTTON_SUPERUSER_FOLLOW_UP", False
)

# make all updates public by default? this will hide the 'is this update
# public' checkbox
HELPDESK_UPDATE_PUBLIC_DEFAULT = getattr(
    settings, "HELPDESK_UPDATE_PUBLIC_DEFAULT", False
)

# only show staff users in ticket owner drop-downs
HELPDESK_STAFF_ONLY_TICKET_OWNERS = getattr(
    settings, "HELPDESK_STAFF_ONLY_TICKET_OWNERS", False
)

# only show staff users in ticket cc drop-down
HELPDESK_STAFF_ONLY_TICKET_CC = getattr(
    settings, "HELPDESK_STAFF_ONLY_TICKET_CC", False
)

# allow the subject to have a configurable template.
HELPDESK_EMAIL_SUBJECT_TEMPLATE = getattr(
    settings,
    "HELPDESK_EMAIL_SUBJECT_TEMPLATE",
    "{{ ticket.ticket }} {{ ticket.title|safe }} %(subject)s",
)
# since django-helpdesk may not work correctly without the ticket ID
# in the subject, let's do a check for it quick:
if HELPDESK_EMAIL_SUBJECT_TEMPLATE.find("ticket.ticket") < 0:
    raise ImproperlyConfigured

# default fallback locale when queue locale not found
HELPDESK_EMAIL_FALLBACK_LOCALE = getattr(
    settings, "HELPDESK_EMAIL_FALLBACK_LOCALE", "en"
)

# default maximum email attachment size, in bytes
# only attachments smaller than this size will be sent via email
HELPDESK_MAX_EMAIL_ATTACHMENT_SIZE = getattr(
    settings, "HELPDESK_MAX_EMAIL_ATTACHMENT_SIZE", 512000
)

# Send email notifications for internal ticket submitter.
HELPDESK_NOTIFY_SUBMITTER_FOR_ALL_TICKET_CHANGES = getattr(
    settings,
    "HELPDESK_NOTIFY_SUBMITTER_FOR_ALL_TICKET_CHANGES",
    False,
)

# If True, private follow-ups (public=False) will not trigger any email notifications
HELPDESK_PRIVATE_FOLLOWUP_MEANS_NO_EMAILS = getattr(
    settings, "HELPDESK_PRIVATE_FOLLOWUP_MEANS_NO_EMAILS", False
)

########################################
# options for staff.create_ticket view #
########################################

# hide the 'assigned to' / 'Case owner' field from the 'create_ticket' view?
HELPDESK_CREATE_TICKET_HIDE_ASSIGNED_TO = getattr(
    settings, "HELPDESK_CREATE_TICKET_HIDE_ASSIGNED_TO", False
)

#################
# email options #
#################

# default Queue email submission settings
QUEUE_EMAIL_BOX_TYPE = getattr(settings, "QUEUE_EMAIL_BOX_TYPE", None)
QUEUE_EMAIL_BOX_SSL = getattr(settings, "QUEUE_EMAIL_BOX_SSL", None)
QUEUE_EMAIL_BOX_HOST = getattr(settings, "QUEUE_EMAIL_BOX_HOST", None)
QUEUE_EMAIL_BOX_USER = getattr(settings, "QUEUE_EMAIL_BOX_USER", None)
QUEUE_EMAIL_BOX_PASSWORD = getattr(settings, "QUEUE_EMAIL_BOX_PASSWORD", None)

# only process emails with a valid tracking ID? (throws away all other mail)
QUEUE_EMAIL_BOX_UPDATE_ONLY = getattr(settings, "QUEUE_EMAIL_BOX_UPDATE_ONLY", False)

# only allow users to access queues that they are members of?
HELPDESK_ENABLE_PER_QUEUE_STAFF_PERMISSION = getattr(
    settings, "HELPDESK_ENABLE_PER_QUEUE_STAFF_PERMISSION", False
)

# use https in the email links
HELPDESK_USE_HTTPS_IN_EMAIL_LINK = getattr(
    settings, "HELPDESK_USE_HTTPS_IN_EMAIL_LINK", settings.SECURE_SSL_REDIRECT
)

# Default to True for backwards compatibility
HELPDESK_TEAMS_MODE_ENABLED = getattr(settings, "HELPDESK_TEAMS_MODE_ENABLED", True)
if HELPDESK_TEAMS_MODE_ENABLED:
    HELPDESK_TEAMS_MODEL = getattr(settings, "HELPDESK_TEAMS_MODEL", "pinax_teams.Team")
    HELPDESK_TEAMS_MIGRATION_DEPENDENCIES = getattr(
        settings,
        "HELPDESK_TEAMS_MIGRATION_DEPENDENCIES",
        [("pinax_teams", "0004_auto_20170511_0856")],
    )
    HELPDESK_KBITEM_TEAM_GETTER = getattr(
        settings, "HELPDESK_KBITEM_TEAM_GETTER", lambda kbitem: kbitem.team
    )
else:
    HELPDESK_TEAMS_MODEL = settings.AUTH_USER_MODEL
    HELPDESK_TEAMS_MIGRATION_DEPENDENCIES = []
    HELPDESK_KBITEM_TEAM_GETTER = lambda _: None  # noqa

# show knowledgebase links?
# If Teams mode is enabled then it has to be on
HELPDESK_KB_ENABLED = (
    True
    if HELPDESK_TEAMS_MODE_ENABLED
    else getattr(settings, "HELPDESK_KB_ENABLED", True)
)

# If set then we always save incoming emails as .eml attachments
# which is quite noisy but very helpful for complicated markup, forwards and so on
# (which gets stripped/corrupted otherwise)
HELPDESK_ALWAYS_SAVE_INCOMING_EMAIL_MESSAGE = getattr(
    settings, "HELPDESK_ALWAYS_SAVE_INCOMING_EMAIL_MESSAGE", False
)

#######################
# email OAUTH         #
#######################

HELPDESK_OAUTH = getattr(
    settings,
    "HELPDESK_OAUTH",
    {"token_url": "", "client_id": "", "secret": "", "scope": [""]},
)

# Set Debug Logging Level for IMAP Services. Default to '0' for No Debugging
HELPDESK_IMAP_DEBUG_LEVEL = getattr(settings, "HELPDESK_IMAP_DEBUG_LEVEL", 0)

#############################################
# file permissions - Attachment directories #
#############################################

# Attachment directories should be created with permission 755 (rwxr-xr-x)
# Override it in your own Django settings.py
HELPDESK_ATTACHMENT_DIR_PERMS = int(
    getattr(settings, "HELPDESK_ATTACHMENT_DIR_PERMS", "755"), 8
)

HELPDESK_VALID_EXTENSIONS = getattr(settings, "VALID_EXTENSIONS", None)
if HELPDESK_VALID_EXTENSIONS:
    # Print to stderr
    print(
        "VALID_EXTENSIONS is deprecated, use HELPDESK_VALID_EXTENSIONS instead",
        file=sys.stderr,
    )
else:
    HELPDESK_VALID_EXTENSIONS = getattr(
        settings,
        "HELPDESK_VALID_EXTENSIONS",
        [
            ".txt",
            ".asc",
            ".htm",
            ".html",
            ".pdf",
            ".doc",
            ".docx",
            ".odt",
            ".jpg",
            ".png",
            ".eml",
        ],
    )

HELPDESK_VALIDATE_ATTACHMENT_TYPES = getattr(
    settings, "HELPDESK_VALIDATE_ATTACHMENT_TYPES", True
)


def get_followup_webhook_urls():
    urls = os.environ.get("HELPDESK_FOLLOWUP_WEBHOOK_URLS", None)
    if urls:
        return re.split(r"[\s],[\s]", urls)


HELPDESK_GET_FOLLOWUP_WEBHOOK_URLS = getattr(
    settings, "HELPDESK_GET_FOLLOWUP_WEBHOOK_URLS", get_followup_webhook_urls
)


def get_new_ticket_webhook_urls():
    urls = os.environ.get("HELPDESK_NEW_TICKET_WEBHOOK_URLS", None)
    if urls:
        return urls.split(",")


HELPDESK_GET_NEW_TICKET_WEBHOOK_URLS = getattr(
    settings, "HELPDESK_GET_NEW_TICKET_WEBHOOK_URLS", get_new_ticket_webhook_urls
)

HELPDESK_WEBHOOK_TIMEOUT = getattr(settings, "HELPDESK_WEBHOOK_TIMEOUT", 3)


LOG_WARN_WHEN_CC_EMAIL_NOT_LINKED_TO_A_USER = getattr(
    settings, "HELPDESK_LOG_WARN_WHEN_CC_EMAIL_NOT_LINKED_TO_A_USER", False
)
LOG_WARN_WHEN_CC_EMAIL_LINKED_TO_MORE_THAN_1_USER = getattr(
    settings, "HELPDESK_LOG_WARN_WHEN_CC_EMAIL_LINKED_TO_MORE_THAN_1_USER", True
)



from datetime import datetime
from django.conf import settings
from django.template import Library
from django.template.defaultfilters import date as date_filter
from helpdesk.settings import (
    CUSTOMFIELD_DATE_FORMAT,
    CUSTOMFIELD_DATETIME_FORMAT,
    CUSTOMFIELD_TIME_FORMAT,
)


register = Library()


@register.filter
def get(value, arg, default=None):
    """Call the dictionary get function"""
    return value.get(arg, default)


@register.filter(expects_localtime=True)
def datetime_string_format(value):
    """
    :param value: String - Expected to be a datetime, date, or time in specific format
    :return: String - reformatted to default datetime, date, or time string if received in one of the expected formats
    """
    try:
        new_value = date_filter(
            datetime.strptime(value, CUSTOMFIELD_DATETIME_FORMAT),
            settings.DATETIME_FORMAT,
        )
    except (TypeError, ValueError):
        try:
            new_value = date_filter(
                datetime.strptime(value, CUSTOMFIELD_DATE_FORMAT), settings.DATE_FORMAT
            )
        except (TypeError, ValueError):
            try:
                new_value = date_filter(
                    datetime.strptime(value, CUSTOMFIELD_TIME_FORMAT),
                    settings.TIME_FORMAT,
                )
            except (TypeError, ValueError):
                # If NoneType return empty string, else return original value
                new_value = "" if value is None else value
    return new_value



# -*- coding: utf-8 -*-
# Generated by Django 1.9.2 on 2016-02-15 21:37
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("helpdesk", "0011_admin_related_improvements"),
    ]

    operations = [
        migrations.AddField(
            model_name="queue",
            name="default_owner",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="default_owner",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Default owner",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0005_queues_no_null"),
    ]

    operations = [
        migrations.AlterField(
            model_name="queue",
            name="email_address",
            field=models.EmailField(
                help_text="All outgoing e-mails for this queue will use this e-mail address. If you use IMAP or POP3, this should be the e-mail address for that mailbox.",
                max_length=254,
                null=True,
                verbose_name="E-Mail Address",
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="ticket",
            name="submitter_email",
            field=models.EmailField(
                help_text="The submitter will receive an email for all public follow-ups left for this task.",
                max_length=254,
                null=True,
                verbose_name="Submitter E-Mail",
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="ticketcc",
            name="email",
            field=models.EmailField(
                help_text="For non-user followers, enter their e-mail address",
                max_length=254,
                null=True,
                verbose_name="E-Mail Address",
                blank=True,
            ),
        ),
    ]



from django.conf import settings
from django.utils.safestring import mark_safe
import logging
import os
from smtplib import SMTPException


logger = logging.getLogger("helpdesk")


def send_templated_mail(
    template_name,
    context,
    recipients,
    sender=None,
    bcc=None,
    fail_silently=False,
    files=None,
    extra_headers=None,
):
    """
    send_templated_mail() is a wrapper around Django's e-mail routines that
    allows us to easily send multipart (text/plain & text/html) e-mails using
    templates that are stored in the database. This lets the admin provide
    both a text and a HTML template for each message.

    template_name is the slug of the template to use for this message (see
        models.EmailTemplate)

    context is a dictionary to be used when rendering the template

    recipients can be either a string, eg 'a@b.com', or a list of strings.

    sender should contain a string, eg 'My Site <me@z.com>'. If you leave it
        blank, it'll use settings.DEFAULT_FROM_EMAIL as a fallback.

    bcc is an optional list of addresses that will receive this message as a
        blind carbon copy.

    fail_silently is passed to Django's mail routine. Set to 'True' to ignore
        any errors at send time.

    files can be a list of tuples. Each tuple should be a filename to attach,
        along with the File objects to be read. files can be blank.

    extra_headers is a dictionary of extra email headers, needed to process
        email replies and keep proper threading.

    """
    from django.core.mail import EmailMultiAlternatives
    from django.template import engines

    from_string = engines["django"].from_string

    from helpdesk.models import EmailTemplate
    from helpdesk.settings import (
        HELPDESK_EMAIL_FALLBACK_LOCALE,
        HELPDESK_EMAIL_SUBJECT_TEMPLATE,
    )

    headers = extra_headers or {}

    locale = context["queue"].get("locale") or HELPDESK_EMAIL_FALLBACK_LOCALE

    try:
        t = EmailTemplate.objects.get(
            template_name__iexact=template_name, locale=locale
        )
    except EmailTemplate.DoesNotExist:
        try:
            t = EmailTemplate.objects.get(
                template_name__iexact=template_name, locale__isnull=True
            )
        except EmailTemplate.DoesNotExist:
            logger.warning('template "%s" does not exist, no mail sent', template_name)
            return  # just ignore if template doesn't exist

    subject_part = (
        from_string(HELPDESK_EMAIL_SUBJECT_TEMPLATE % {"subject": t.subject})
        .render(context)
        .replace("\n", "")
        .replace("\r", "")
    )

    footer_file = os.path.join("helpdesk", locale, "email_text_footer.txt")

    text_part = from_string(
        "%s\n\n{%% include '%s' %%}" % (t.plain_text, footer_file)
    ).render(context)

    email_html_base_file = os.path.join("helpdesk", locale, "email_html_base.html")
    # keep new lines in html emails
    if "comment" in context:
        context["comment"] = mark_safe(context["comment"].replace("\r\n", "<br>"))

    html_part = from_string(
        "{%% extends '%s' %%}"
        "{%% block title %%}%s{%% endblock %%}"
        "{%% block content %%}%s{%% endblock %%}"
        % (email_html_base_file, t.heading, t.html)
    ).render(context)

    if isinstance(recipients, str):
        if recipients.find(","):
            recipients = recipients.split(",")
    elif type(recipients) is not list:
        recipients = [recipients]

    msg = EmailMultiAlternatives(
        subject_part,
        text_part,
        sender or settings.DEFAULT_FROM_EMAIL,
        recipients,
        bcc=bcc,
        headers=headers,
    )
    msg.attach_alternative(html_part, "text/html")

    if files:
        for filename, filefield in files:
            filefield.open("rb")
            content = filefield.read()
            msg.attach(filename, content)
            filefield.close()
    logger.debug("Sending email to: {!r}".format(recipients))

    try:
        return msg.send()
    except SMTPException as e:
        logger.exception(
            "SMTPException raised while sending email to {}".format(recipients)
        )
        if not fail_silently:
            raise e
        return 0



import typing

from django.core.exceptions import ValidationError
from django.contrib.auth import get_user_model
from django.utils import timezone
from django.utils.translation import gettext as _

from helpdesk.lib import safe_template_context
from helpdesk import settings as helpdesk_settings
from helpdesk.lib import process_attachments
from helpdesk.decorators import (
    is_helpdesk_staff,
)
from helpdesk.models import (
    FollowUp,
    Ticket,
    TicketCC,
)
from helpdesk.signals import update_ticket_done

User = get_user_model()


def add_staff_subscription(user: User, ticket: Ticket) -> None:
    """Auto subscribe the staff member if that's what the settigs say and the
    user is authenticated and a staff member"""
    if (
        helpdesk_settings.HELPDESK_AUTO_SUBSCRIBE_ON_TICKET_RESPONSE
        and user.is_authenticated
        and return_ticketccstring_and_show_subscribe(user, ticket)[1]
    ):
        subscribe_to_ticket_updates(ticket, user.id)


def return_ticketccstring_and_show_subscribe(user, ticket):
    """used in view_ticket() and followup_edit()"""
    # create the ticketcc_string and check whether current user is already
    # subscribed
    username = user.get_username().upper()
    try:
        useremail = user.email.upper()
    except AttributeError:
        useremail = ""
    strings_to_check = list()
    strings_to_check.append(username)
    strings_to_check.append(useremail)

    ticketcc_string = ""
    all_ticketcc = ticket.ticketcc_set.all()
    counter_all_ticketcc = len(all_ticketcc) - 1
    show_subscribe = True
    for i, ticketcc in enumerate(all_ticketcc):
        ticketcc_this_entry = str(ticketcc.display)
        ticketcc_string += ticketcc_this_entry
        if i < counter_all_ticketcc:
            ticketcc_string += ", "
        if strings_to_check.__contains__(ticketcc_this_entry.upper()):
            show_subscribe = False

    # check whether current user is a submitter or assigned to ticket
    assignedto_username = str(ticket.assigned_to).upper()
    strings_to_check = list()
    if ticket.submitter_email is not None:
        submitter_email = ticket.submitter_email.upper()
        strings_to_check.append(submitter_email)
    strings_to_check.append(assignedto_username)
    if strings_to_check.__contains__(username) or strings_to_check.__contains__(
        useremail
    ):
        show_subscribe = False

    return ticketcc_string, show_subscribe


def subscribe_to_ticket_updates(
    ticket, user_id=None, email=None, can_view=True, can_update=False
):
    if ticket is not None:
        queryset = TicketCC.objects.filter(ticket=ticket, user_id=user_id, email=email)

        # Don't create duplicate entries for subscribers
        if queryset.count() > 0:
            return queryset.first()

        if user_id is None and len(email) < 5:
            raise ValidationError(
                _(
                    "When you add somebody on Cc, you must provide either a User or a valid email. Email: %s"
                    % email
                )
            )

        return ticket.ticketcc_set.create(
            user_id=user_id, email=email, can_view=can_view, can_update=can_update
        )


def get_and_set_ticket_status(
    new_status: int, ticket: Ticket, follow_up: FollowUp
) -> typing.Tuple[str, int]:
    """Performs comparision on previous status to new status,
    updating the title as required.

    Returns:
        The old status as a display string, old status code string
    """
    old_status_str = ticket.get_status_display()
    old_status = ticket.status
    if new_status != ticket.status:
        ticket.status = new_status
        ticket.save()
        follow_up.new_status = new_status
        if follow_up.title:
            follow_up.title += " and %s" % ticket.get_status_display()
        else:
            follow_up.title = "%s" % ticket.get_status_display()

    if not follow_up.title:
        if follow_up.comment:
            follow_up.title = _("Comment")
        else:
            follow_up.title = _("Updated")

    follow_up.save()
    return old_status_str, old_status


def process_email_notifications_for_ticket_update(
    public: bool,
    ticket: Ticket,
    follow_up: FollowUp,
    context: dict,
    messages_sent_to: typing.Set[str],
    files: typing.List[typing.Tuple[str, str]],
    reassigned: bool = False,
):
    """
    Sends email notifications when the ticket is updated in any way.
    """
    # If the setting is enabled and this is a private follow-up, skip all email notifications
    if helpdesk_settings.HELPDESK_PRIVATE_FOLLOWUP_MEANS_NO_EMAILS and not public:
        return

    template_prefix = get_email_template_prefix(reassigned, follow_up)
    if helpdesk_settings.HELPDESK_NOTIFY_SUBMITTER_FOR_ALL_TICKET_CHANGES or (
        public
        and (
            follow_up.comment
            or (follow_up.new_status in (Ticket.RESOLVED_STATUS, Ticket.CLOSED_STATUS))
        )
    ):
        # Use hard coded prefix for submitter updates on  tickets for backwards compatibility
        # TODO: possibly make the template prefix modification configurable
        messages_sent_to.update(
            ticket.send(
                {
                    "submitter": ("updated_submitter", context),
                },
                dont_send_to=messages_sent_to,
                fail_silently=True,
                files=files,
            )
        )
    if ticket.assigned_to and (
        ticket.assigned_to.usersettings_helpdesk.email_on_ticket_change
        or (
            reassigned
            and ticket.assigned_to.usersettings_helpdesk.email_on_ticket_assign
        )
    ):
        messages_sent_to.update(
            ticket.send(
                {"assigned_to": (template_prefix + "owner", context)},
                dont_send_to=messages_sent_to,
                fail_silently=True,
                files=files,
            )
        )

    messages_sent_to.update(
        ticket.send(
            {"ticket_cc": (template_prefix + "cc", context)},
            dont_send_to=messages_sent_to,
            fail_silently=True,
            files=files,
        )
    )


def get_email_template_prefix(reassigned, follow_up: FollowUp) -> str:
    if reassigned:
        return "assigned_"
    elif follow_up.new_status == Ticket.RESOLVED_STATUS:
        return "resolved_"
    elif follow_up.new_status == Ticket.CLOSED_STATUS:
        return "closed_"
    else:
        return "updated_"


def update_ticket(
    user,
    ticket,
    title=None,
    comment="",
    files=None,
    public=False,
    owner=-1,
    priority=-1,
    queue=-1,
    new_status=None,
    time_spent=None,
    due_date=None,
    new_checklists=None,
    message_id=None,
    customfields_form=None,
):
    # We need to allow the 'ticket' and 'queue' contexts to be applied to the
    # comment.
    context = safe_template_context(ticket)
    if title is None:
        title = ticket.title
    if priority == -1:
        priority = ticket.priority
    if queue == -1:
        queue = ticket.queue.id
    if new_status is None:
        new_status = ticket.status
    if new_checklists is None:
        new_checklists = {}

    from django.template import engines

    template_func = engines["django"].from_string
    # this prevents system from trying to render any template tags
    # broken into two stages to prevent changes from first replace being themselves
    # changed by the second replace due to conflicting syntax
    comment = comment.replace("{%", "X-HELPDESK-COMMENT-VERBATIM").replace(
        "%}", "X-HELPDESK-COMMENT-ENDVERBATIM"
    )
    comment = comment.replace(
        "X-HELPDESK-COMMENT-VERBATIM", "{% verbatim %}{%"
    ).replace("X-HELPDESK-COMMENT-ENDVERBATIM", "%}{% endverbatim %}")
    # render the neutralized template
    comment = template_func(comment).render(context)

    if owner == -1 and ticket.assigned_to:
        owner = ticket.assigned_to.id

    f = FollowUp(
        ticket=ticket,
        date=timezone.now(),
        comment=comment,
        time_spent=time_spent,
        message_id=message_id,
        title=title,
    )

    if is_helpdesk_staff(user):
        f.user = user

    f.public = public

    reassigned = False

    old_owner = ticket.assigned_to
    if owner != -1:
        if owner != 0 and (
            (ticket.assigned_to and owner != ticket.assigned_to.id)
            or not ticket.assigned_to
        ):
            new_user = User.objects.get(id=owner)
            f.title = _("Assigned to %(username)s") % {
                "username": new_user.get_username(),
            }
            ticket.assigned_to = new_user
            reassigned = True
        # user changed owner to 'unassign'
        elif owner == 0 and ticket.assigned_to is not None:
            f.title = _("Unassigned")
            ticket.assigned_to = None

    old_status_str, old_status = get_and_set_ticket_status(new_status, ticket, f)

    files = process_attachments(f, files) if files else []

    if title and title != ticket.title:
        f.ticketchange_set.create(
            field=_("Title"),
            old_value=ticket.title,
            new_value=title,
        )
        ticket.title = title

    if new_status != old_status:
        f.ticketchange_set.create(
            field=_("Status"),
            old_value=old_status_str,
            new_value=ticket.get_status_display(),
        )

    if ticket.assigned_to != old_owner:
        f.ticketchange_set.create(
            field=_("Owner"),
            old_value=old_owner,
            new_value=ticket.assigned_to if ticket.assigned_to else _("Unassigned"),
        )

    if priority != ticket.priority:
        f.ticketchange_set.create(
            field=_("Priority"),
            old_value=ticket.priority,
            new_value=priority,
        )
        ticket.priority = priority

    if queue != ticket.queue.id:
        f.ticketchange_set.create(
            field=_("Queue"),
            old_value=ticket.queue.id,
            new_value=queue,
        )
        ticket.queue_id = queue

    if due_date and due_date != ticket.due_date:
        f.ticketchange_set.create(
            field=_("Due on"),
            old_value=ticket.due_date,
            new_value=due_date,
        )
        ticket.due_date = due_date

    # save custom fields and ticket changes
    if customfields_form and customfields_form.is_valid():
        customfields_form.save(followup=f)

    for checklist in ticket.checklists.all():
        if checklist.id not in new_checklists:
            continue
        new_completed_tasks = new_checklists[checklist.id]
        for task in checklist.tasks.all():
            changed = None

            # Add completion if it was not done yet
            if not task.completion_date and task.id in new_completed_tasks:
                task.completion_date = timezone.now()
                changed = "completed"
            # Remove it if it was done before
            elif task.completion_date and task.id not in new_completed_tasks:
                task.completion_date = None
                changed = "uncompleted"

            # Save and add ticket change if task state has changed
            if changed:
                task.save(update_fields=["completion_date"])
                f.ticketchange_set.create(
                    field=f"[{checklist.name}] {task.description}",
                    old_value=_("To do") if changed == "completed" else _("Completed"),
                    new_value=_("Completed") if changed == "completed" else _("To do"),
                )

    if new_status in (Ticket.RESOLVED_STATUS, Ticket.CLOSED_STATUS) and (
        new_status == Ticket.RESOLVED_STATUS or ticket.resolution is None
    ):
        ticket.resolution = comment

    # ticket might have changed above, so we re-instantiate context with the
    # (possibly) updated ticket.
    context = safe_template_context(ticket)
    context.update(
        resolution=ticket.resolution,
        comment=f.comment,
    )

    messages_sent_to = set()
    try:
        messages_sent_to.add(user.email)
    except AttributeError:
        pass
    process_email_notifications_for_ticket_update(
        public, ticket, f, context, messages_sent_to, files, reassigned=reassigned
    )
    ticket.save()

    # emit signal with followup when the ticket update is done
    # internally used for webhooks
    update_ticket_done.send(sender="update_ticket", followup=f)

    # auto subscribe user if enabled
    add_staff_subscription(user, ticket)
    return f



"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008 Jutda. All Rights Reserved. See LICENSE for details.

views/kb.py - Public-facing knowledgebase views. The knowledgebase is a
              simple categorised question/answer system to show common
              resolutions to common problems.
"""

from django.http import Http404, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.views.decorators.clickjacking import xframe_options_exempt
from helpdesk import settings as helpdesk_settings, user
from helpdesk.models import KBCategory, KBItem


def index(request):
    huser = user.huser_from_request(request)
    # TODO: It'd be great to have a list of most popular items here.
    return render(
        request,
        "helpdesk/kb_index.html",
        {
            "kb_categories": huser.get_allowed_kb_categories(),
            "helpdesk_settings": helpdesk_settings,
        },
    )


def category(request, slug, iframe=False):
    category = get_object_or_404(KBCategory, slug__iexact=slug)
    if not user.huser_from_request(request).can_access_kbcategory(category):
        raise Http404
    items = category.kbitem_set.filter(enabled=True)
    selected_item = request.GET.get("kbitem", None)
    try:
        selected_item = int(selected_item)
    except TypeError:
        pass
    qparams = request.GET.copy()
    try:
        del qparams["kbitem"]
    except KeyError:
        pass
    template = "helpdesk/kb_category.html"
    if iframe:
        template = "helpdesk/kb_category_iframe.html"
    staff = request.user.is_authenticated and request.user.is_staff
    return render(
        request,
        template,
        {
            "category": category,
            "items": items,
            "selected_item": selected_item,
            "query_param_string": qparams.urlencode(),
            "helpdesk_settings": helpdesk_settings,
            "iframe": iframe,
            "staff": staff,
        },
    )


@xframe_options_exempt
def category_iframe(request, slug):
    return category(request, slug, iframe=True)


def vote(request, item, vote):
    item = get_object_or_404(KBItem, pk=item)
    if request.method == "POST":
        if vote == "up":
            if not item.voted_by.filter(pk=request.user.pk):
                item.votes += 1
                item.voted_by.add(request.user.pk)
                item.recommendations += 1
            if item.downvoted_by.filter(pk=request.user.pk):
                item.votes -= 1
                item.downvoted_by.remove(request.user.pk)
        if vote == "down":
            if not item.downvoted_by.filter(pk=request.user.pk):
                item.votes += 1
                item.downvoted_by.add(request.user.pk)
                item.recommendations -= 1
            if item.voted_by.filter(pk=request.user.pk):
                item.votes -= 1
                item.voted_by.remove(request.user.pk)
        item.save()
    return HttpResponseRedirect(item.get_absolute_url())



#!/usr/bin/python
"""
Jutda Helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008 Jutda. All Rights Reserved. See LICENSE for details.

scripts/get_email.py - Designed to be run from cron, this script checks the
                       POP and IMAP boxes, or a local mailbox directory,
                       defined for the queues within a
                       helpdesk, creating tickets from the new messages (or
                       adding to existing tickets if needed)
"""

from django.core.management.base import BaseCommand
from helpdesk.email import process_email


class Command(BaseCommand):
    help = (
        "Process django-helpdesk queues and process e-mails via POP3/IMAP or "
        "from a local mailbox directory as required, feeding them into the helpdesk."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--quiet",
            action="store_true",
            dest="quiet",
            default=False,
            help="Hide details about each queue/message as they are processed",
        )
        parser.add_argument(
            "--debug_to_stdout",
            action="store_true",
            dest="debug_to_stdout",
            default=False,
            help="Log additional messaging to stdout.",
        )

    def handle(self, *args, **options):
        quiet = options.get("quiet")
        debug_to_stdout = options.get("debug_to_stdout")
        process_email(quiet=quiet, debug_to_stdout=debug_to_stdout)


if __name__ == "__main__":
    process_email()



# Generated by Django 2.2.13 on 2020-10-27 17:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0032_kbitem_enabled"),
    ]

    operations = [
        migrations.AddField(
            model_name="ticket",
            name="merged_to",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="merged_tickets",
                to="helpdesk.Ticket",
                verbose_name="merged to",
            ),
        ),
    ]



# Generated by Django 3.2.7 on 2021-10-05 10:21

from django.db import migrations, models
import helpdesk.models
import helpdesk.validators


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0035_alter_email_on_ticket_change"),
    ]

    operations = [
        migrations.AlterField(
            model_name="followupattachment",
            name="file",
            field=models.FileField(
                max_length=1000,
                upload_to=helpdesk.models.attachment_path,
                validators=[helpdesk.validators.validate_file_extension],
                verbose_name="File",
            ),
        ),
        migrations.AlterField(
            model_name="kbiattachment",
            name="file",
            field=models.FileField(
                max_length=1000,
                upload_to=helpdesk.models.attachment_path,
                validators=[helpdesk.validators.validate_file_extension],
                verbose_name="File",
            ),
        ),
    ]



# Generated by Django 2.0.7 on 2018-10-19 14:11
from django.db import migrations, models
import helpdesk.models


try:
    import pickle
except ImportError:
    import cPickle as pickle  # pyright: ignore[reportMissingImports]
try:
    # Python 2 support
    from base64 import urlsafe_b64decode as b64decode
except ImportError:
    # Python 3 support
    from base64 import decodebytes as b64decode


# return a python dictionary representing the pickled data.
def unpickle_settings(settings_pickled):
    try:
        return pickle.loads(b64decode(settings_pickled.encode("utf-8")))
    except Exception:
        return {}


def move_old_values(apps, schema_editor):
    UserSettings = apps.get_model("helpdesk", "UserSettings")
    db_alias = schema_editor.connection.alias

    for user_settings in UserSettings.objects.using(db_alias).all():
        if user_settings.settings_pickled:
            settings_dict = unpickle_settings(user_settings.settings_pickled)
            for setting, value in settings_dict.items():
                user_settings.__setattr__(setting, value)


def pickle_settings(apps, schema_editor):
    """
    Pickle settings back to the old format. This is necessary for reversing this migration.
    """
    UserSettings = apps.get_model("helpdesk", "UserSettings")
    db_alias = schema_editor.connection.alias
    for user_settings in UserSettings.objects.using(db_alias).all():
        settings_dict = {
            "email_on_ticket_assign": user_settings.email_on_ticket_assign,
            "email_on_ticket_change": user_settings.email_on_ticket_change,
            "login_view_ticketlist": user_settings.login_view_ticketlist,
            "tickets_per_page": user_settings.tickets_per_page,
            "use_email_as_submitter": user_settings.use_email_as_submitter,
        }
        user_settings.settings_pickled = pickle.dumps(settings_dict)
        user_settings.save()


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0019_ticket_secret_key"),
    ]

    operations = [
        migrations.AddField(
            model_name="usersettings",
            name="email_on_ticket_assign",
            field=models.BooleanField(
                default=helpdesk.models.email_on_ticket_assign_default,
                help_text="If you are assigned a ticket via the web, do you want to receive an e-mail?",
                verbose_name="E-mail me when assigned a ticket?",
            ),
        ),
        migrations.AddField(
            model_name="usersettings",
            name="email_on_ticket_change",
            field=models.BooleanField(
                default=helpdesk.models.email_on_ticket_change_default,
                help_text="If you're the ticket owner and the ticket is changed via the web by somebody else, "
                "do you want to receive an e-mail?",
                verbose_name="E-mail me on ticket change?",
            ),
        ),
        migrations.AddField(
            model_name="usersettings",
            name="login_view_ticketlist",
            field=models.BooleanField(
                default=helpdesk.models.login_view_ticketlist_default,
                help_text="Display the ticket list upon login? Otherwise, the dashboard is shown.",
                verbose_name="Show Ticket List on Login?",
            ),
        ),
        migrations.AddField(
            model_name="usersettings",
            name="tickets_per_page",
            field=models.IntegerField(
                choices=[(10, "10"), (25, "25"), (50, "50"), (100, "100")],
                default=helpdesk.models.tickets_per_page_default,
                help_text="How many tickets do you want to see on the Ticket List page?",
                verbose_name="Number of tickets to show per page",
            ),
        ),
        migrations.AddField(
            model_name="usersettings",
            name="use_email_as_submitter",
            field=models.BooleanField(
                default=helpdesk.models.use_email_as_submitter_default,
                help_text="When you submit a ticket, do you want to automatically use your e-mail address as the "
                "submitter address? You can type a different e-mail address when entering the ticket if needed, "
                "this option only changes the default.",
                verbose_name="Use my e-mail address when submitting tickets?",
            ),
        ),
        migrations.AlterField(
            model_name="usersettings",
            name="settings_pickled",
            field=models.TextField(
                blank=True,
                help_text="DEPRECATED! This is a base64-encoded representation of a pickled Python dictionary. Do not "
                "change this field via the admin.",
                null=True,
                verbose_name="DEPRECATED! Settings Dictionary DEPRECATED!",
            ),
        ),
        migrations.RunPython(move_old_values, reverse_code=pickle_settings),
    ]



"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008 Jutda. All Rights Reserved. See LICENSE for details.

urls.py - Mapping of URL's to our various views. Note we always used NAMED
          views for simplicity in linking later on.
"""

from django.contrib.auth import views as auth_views
from django.contrib.auth.decorators import login_required
from django.urls import include, path, re_path
from django.views.generic import TemplateView
from helpdesk import settings as helpdesk_settings
from helpdesk.decorators import helpdesk_staff_member_required, protect_view
from helpdesk.views import feeds, login, public, staff
from helpdesk.views.api import (
    CreateUserView,
    FollowUpAttachmentViewSet,
    FollowUpViewSet,
    TicketViewSet,
    UserTicketViewSet,
)
from rest_framework.routers import DefaultRouter


if helpdesk_settings.HELPDESK_KB_ENABLED:
    from helpdesk.views import kb

try:
    # TODO: why is it imported? due to some side-effect or by mistake?
    import helpdesk.tasks  # NOQA
except ImportError:
    pass


class DirectTemplateView(TemplateView):
    extra_context = None

    def get_context_data(self, **kwargs):
        context = super(self.__class__, self).get_context_data(**kwargs)
        if self.extra_context is not None:
            for key, value in self.extra_context.items():
                if callable(value):
                    context[key] = value()
                else:
                    context[key] = value
        return context


app_name = "helpdesk"

base64_pattern = r"(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$"

urlpatterns = [
    path("dashboard/", staff.dashboard, name="dashboard"),
    path("tickets/", staff.ticket_list, name="list"),
    path("tickets/update/", staff.mass_update, name="mass_update"),
    path("tickets/merge", staff.merge_tickets, name="merge_tickets"),
    path("tickets/<int:ticket_id>/", staff.view_ticket, name="view"),
    path(
        "tickets/<int:ticket_id>/followup_edit/<int:followup_id>/",
        staff.followup_edit,
        name="followup_edit",
    ),
    path(
        "tickets/<int:ticket_id>/followup_delete/<int:followup_id>/",
        staff.followup_delete,
        name="followup_delete",
    ),
    path("tickets/<int:ticket_id>/edit/", staff.edit_ticket, name="edit"),
    path("tickets/<int:ticket_id>/update/", staff.update_ticket_view, name="update"),
    path("tickets/<int:ticket_id>/delete/", staff.delete_ticket, name="delete"),
    path("tickets/<int:ticket_id>/hold/", staff.hold_ticket, name="hold"),
    path("tickets/<int:ticket_id>/unhold/", staff.unhold_ticket, name="unhold"),
    path("tickets/<int:ticket_id>/cc/", staff.ticket_cc, name="ticket_cc"),
    path("tickets/<int:ticket_id>/cc/add/", staff.ticket_cc_add, name="ticket_cc_add"),
    path(
        "tickets/<int:ticket_id>/cc/delete/<int:cc_id>/",
        staff.ticket_cc_del,
        name="ticket_cc_del",
    ),
    path(
        "tickets/<int:ticket_id>/dependency/add/",
        staff.ticket_dependency_add,
        name="ticket_dependency_add",
    ),
    path(
        "tickets/<int:ticket_id>/dependency/delete/<int:dependency_id>/",
        staff.ticket_dependency_del,
        name="ticket_dependency_del",
    ),
    path(
        "tickets/<int:ticket_id>/resolves/add/",
        staff.ticket_resolves_add,
        name="ticket_resolves_add",
    ),
    path(
        "tickets/<int:ticket_id>/resolves/delete/<int:dependency_id>/",
        staff.ticket_resolves_del,
        name="ticket_resolves_del",
    ),
    path(
        "tickets/<int:ticket_id>/attachment_delete/<int:attachment_id>/",
        staff.attachment_del,
        name="attachment_del",
    ),
    path(
        "tickets/<int:ticket_id>/checklists/<int:checklist_id>/",
        staff.edit_ticket_checklist,
        name="edit_ticket_checklist",
    ),
    path(
        "tickets/<int:ticket_id>/checklists/<int:checklist_id>/delete/",
        staff.delete_ticket_checklist,
        name="delete_ticket_checklist",
    ),
    re_path(r"^raw/(?P<type_>\w+)/$", staff.raw_details, name="raw"),
    path("rss/", staff.rss_list, name="rss_index"),
    path("reports/", staff.report_index, name="report_index"),
    re_path(r"^reports/(?P<report>\w+)/$", staff.run_report, name="run_report"),
    path("saved-searches/", staff.saved_searches_list, name="saved_searches_list"),
    path("save_query/", staff.save_query, name="savequery"),
    path("delete_query/<int:pk>/", staff.delete_saved_query, name="delete_query"),
    path("settings/", staff.EditUserSettingsView.as_view(), name="user_settings"),
    path("ignore/", staff.email_ignore, name="email_ignore"),
    path("ignore/add/", staff.email_ignore_add, name="email_ignore_add"),
    path("ignore/delete/<int:id>/", staff.email_ignore_del, name="email_ignore_del"),
    path("checklist-templates/", staff.checklist_templates, name="checklist_templates"),
    path(
        "checklist-templates/<int:checklist_template_id>/",
        staff.checklist_templates,
        name="edit_checklist_template",
    ),
    path(
        "checklist-templates/<int:checklist_template_id>/delete/",
        staff.delete_checklist_template,
        name="delete_checklist_template",
    ),
    re_path(
        r"^datatables_ticket_list/(?P<query>{})$".format(base64_pattern),
        staff.datatables_ticket_list,
        name="datatables_ticket_list",
    ),
    re_path(
        r"^timeline_ticket_list/(?P<query>{})$".format(base64_pattern),
        staff.timeline_ticket_list,
        name="timeline_ticket_list",
    ),
]

if helpdesk_settings.HELPDESK_ENABLE_DEPENDENCIES_ON_TICKET:
    urlpatterns += [
        path(
            "tickets/<int:ticket_id>/dependency/add/",
            staff.ticket_dependency_add,
            name="ticket_dependency_add",
        ),
        path(
            "tickets/<int:ticket_id>/dependency/delete/<int:dependency_id>/",
            staff.ticket_dependency_del,
            name="ticket_dependency_del",
        ),
    ]

urlpatterns += [
    path("", protect_view(public.Homepage.as_view()), name="home"),
    path(
        "tickets/my-tickets/",
        protect_view(public.MyTickets.as_view()),
        name="my-tickets",
    ),
    path("tickets/submit/", public.create_ticket, name="submit"),
    path(
        "tickets/submit_iframe/",
        protect_view(public.CreateTicketIframeView.as_view()),
        name="submit_iframe",
    ),
    path(
        "tickets/success_iframe/",  # Ticket was submitted successfully
        protect_view(public.SuccessIframeView.as_view()),
        name="success_iframe",
    ),
    path("view/", protect_view(public.ViewTicket.as_view()), name="public_view"),
    path("change_language/", public.change_language, name="public_change_language"),
]

urlpatterns += [
    re_path(
        r"^rss/user/(?P<user_name>[^/]+)/",
        helpdesk_staff_member_required(feeds.OpenTicketsByUser()),
        name="rss_user",
    ),
    re_path(
        r"^rss/user/(?P<user_name>[^/]+)/(?P<queue_slug>[A-Za-z0-9_-]+)/$",
        helpdesk_staff_member_required(feeds.OpenTicketsByUser()),
        name="rss_user_queue",
    ),
    re_path(
        r"^rss/queue/(?P<queue_slug>[A-Za-z0-9_-]+)/$",
        helpdesk_staff_member_required(feeds.OpenTicketsByQueue()),
        name="rss_queue",
    ),
    path(
        "rss/unassigned/",
        helpdesk_staff_member_required(feeds.UnassignedTickets()),
        name="rss_unassigned",
    ),
    path(
        "rss/recent_activity/",
        helpdesk_staff_member_required(feeds.RecentFollowUps()),
        name="rss_activity",
    ),
]


router = DefaultRouter()
router.register(r"tickets", TicketViewSet, basename="ticket")
router.register(r"user_tickets", UserTicketViewSet, basename="user_tickets")
router.register(r"followups", FollowUpViewSet, basename="followups")
router.register(
    r"followups-attachments", FollowUpAttachmentViewSet, basename="followupattachments"
)
router.register(r"users", CreateUserView, basename="user")
urlpatterns += [path("api/", include(router.urls))]


urlpatterns += [
    path("login/", login.login, name="login"),
    path(
        "logout/",
        auth_views.LogoutView.as_view(
            template_name="helpdesk/registration/logged_out.html"
        ),
        name="logout",
    ),
    path(
        "password_change/",
        auth_views.PasswordChangeView.as_view(
            template_name="helpdesk/registration/change_password.html",
            success_url="./done",
        ),
        name="password_change",
    ),
    path(
        "password_change/done",
        auth_views.PasswordChangeDoneView.as_view(
            template_name="helpdesk/registration/change_password_done.html",
        ),
        name="password_change_done",
    ),
]

if helpdesk_settings.HELPDESK_KB_ENABLED:
    urlpatterns += [
        path("kb/", kb.index, name="kb_index"),
        re_path(r"^kb/(?P<slug>[A-Za-z0-9_-]+)/$", kb.category, name="kb_category"),
        re_path(r"^kb/(?P<item>\d+)/vote/(?P<vote>up|down)/$", kb.vote, name="kb_vote"),
        re_path(
            r"^kb_iframe/(?P<slug>[A-Za-z0-9_-]+)/$",
            kb.category_iframe,
            name="kb_category_iframe",
        ),
    ]

urlpatterns += [
    path(
        "help/context/",
        TemplateView.as_view(template_name="helpdesk/help_context.html"),
        name="help_context",
    ),
    path(
        "system_settings/",
        login_required(
            DirectTemplateView.as_view(template_name="helpdesk/system_settings.html")
        ),
        name="system_settings",
    ),
]



# Generated by Django 2.0.1 on 2018-09-07 21:22
from django.db import migrations, models
import helpdesk.models


def clear_secret_keys(apps, schema_editor):
    Ticket = apps.get_model("helpdesk", "Ticket")
    db_alias = schema_editor.connection.alias

    for ticket in Ticket.objects.using(db_alias).all():
        ticket.secret_key = ""
        ticket.save()


def no_op(apps, schema_editor):
    """
    This is added only to make migrations reversible.
    """
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0018_ticket_secret_key"),
    ]

    operations = [
        migrations.AddField(
            model_name="ticket",
            name="secret_key",
            field=models.CharField(
                default=helpdesk.models.mk_secret,
                max_length=36,
                verbose_name="Secret key needed for viewing/editing ticket by non-logged in users",
            ),
        ),
        migrations.RunPython(clear_secret_keys, reverse_code=no_op),
    ]



from django.apps import AppConfig


class HelpdeskConfig(AppConfig):
    name = "helpdesk"
    verbose_name = "Helpdesk"
    # for Django 3.2 support:
    # see:
    # https://docs.djangoproject.com/en/3.2/ref/applications/#django.apps.AppConfig.default_auto_field
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from . import webhooks  # noqa: F401



# Generated by Django 2.2.10 on 2020-02-25 13:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0031_auto_20200225_1440"),
    ]

    operations = [
        migrations.AddField(
            model_name="kbitem",
            name="enabled",
            field=models.BooleanField(
                default=True, verbose_name="Enabled to display to users"
            ),
        ),
    ]



"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008 Jutda. All Rights Reserved. See LICENSE for details.

views/feeds.py - A handful of staff-only RSS feeds to provide ticket details
                 to feed readers or similar software.
"""

from django.contrib.auth import get_user_model
from django.contrib.syndication.views import Feed
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext as _
from helpdesk.models import FollowUp, Queue, Ticket


User = get_user_model()


Q_OPEN_STATUSES = Q()
for open_status in Ticket.OPEN_STATUSES:
    Q_OPEN_STATUSES |= Q(status=open_status)


class OpenTicketsByUser(Feed):
    title_template = "helpdesk/rss/ticket_title.html"
    description_template = "helpdesk/rss/ticket_description.html"

    def get_object(self, request, user_name, queue_slug=None):
        user = get_object_or_404(User, username=user_name)
        if queue_slug:
            queue = get_object_or_404(Queue, slug=queue_slug)
        else:
            queue = None

        return {"user": user, "queue": queue}

    def title(self, obj):
        if obj["queue"]:
            return _("Helpdesk: Open Tickets in queue %(queue)s for %(username)s") % {
                "queue": obj["queue"].title,
                "username": obj["user"].get_username(),
            }
        else:
            return _("Helpdesk: Open Tickets for %(username)s") % {
                "username": obj["user"].get_username(),
            }

    def description(self, obj):
        if obj["queue"]:
            return _(
                "Open and Reopened Tickets in queue %(queue)s for %(username)s"
            ) % {
                "queue": obj["queue"].title,
                "username": obj["user"].get_username(),
            }
        else:
            return _("Open and Reopened Tickets for %(username)s") % {
                "username": obj["user"].get_username(),
            }

    def link(self, obj):
        if obj["queue"]:
            return "%s?assigned_to=%s&queue=%s" % (
                reverse("helpdesk:list"),
                obj["user"].id,
                obj["queue"].id,
            )
        else:
            return "%s?assigned_to=%s" % (
                reverse("helpdesk:list"),
                obj["user"].id,
            )

    def items(self, obj):
        if obj["queue"]:
            return (
                Ticket.objects.filter(assigned_to=obj["user"])
                .filter(queue=obj["queue"])
                .filter(Q_OPEN_STATUSES)
            )
        else:
            return Ticket.objects.filter(assigned_to=obj["user"]).filter(
                Q_OPEN_STATUSES
            )

    def item_pubdate(self, item):
        return item.created

    def item_author_name(self, item):
        if item.assigned_to:
            return item.assigned_to.get_username()
        else:
            return _("Unassigned")


class UnassignedTickets(Feed):
    title_template = "helpdesk/rss/ticket_title.html"
    description_template = "helpdesk/rss/ticket_description.html"

    title = _("Helpdesk: Unassigned Tickets")
    description = _("Unassigned Open and Reopened tickets")
    link = ""  # '%s?assigned_to=' % reverse('helpdesk:list')

    def items(self, obj):
        return Ticket.objects.filter(assigned_to__isnull=True).filter(Q_OPEN_STATUSES)

    def item_pubdate(self, item):
        return item.created

    def item_author_name(self, item):
        if item.assigned_to:
            return item.assigned_to.get_username()
        else:
            return _("Unassigned")


class RecentFollowUps(Feed):
    title_template = "helpdesk/rss/recent_activity_title.html"
    description_template = "helpdesk/rss/recent_activity_description.html"

    title = _("Helpdesk: Recent Followups")
    description = _(
        "Recent FollowUps, such as e-mail replies, comments, attachments and resolutions"
    )
    link = "/tickets/"  # reverse('helpdesk:list')

    def items(self):
        return FollowUp.objects.order_by("-date")[:20]


class OpenTicketsByQueue(Feed):
    title_template = "helpdesk/rss/ticket_title.html"
    description_template = "helpdesk/rss/ticket_description.html"

    def get_object(self, request, queue_slug):
        return get_object_or_404(Queue, slug=queue_slug)

    def title(self, obj):
        return _("Helpdesk: Open Tickets in queue %(queue)s") % {
            "queue": obj.title,
        }

    def description(self, obj):
        return _("Open and Reopened Tickets in queue %(queue)s") % {
            "queue": obj.title,
        }

    def link(self, obj):
        return "%s?queue=%s" % (
            reverse("helpdesk:list"),
            obj.id,
        )

    def items(self, obj):
        return Ticket.objects.filter(queue=obj).filter(Q_OPEN_STATUSES)

    def item_pubdate(self, item):
        return item.created

    def item_author_name(self, item):
        if item.assigned_to:
            return item.assigned_to.get_username()
        else:
            return _("Unassigned")



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-14 23:47
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0012_queue_default_owner"),
    ]

    operations = [
        migrations.AddField(
            model_name="queue",
            name="email_box_local_dir",
            field=models.CharField(
                blank=True,
                help_text="If using a local directory, what directory path do you wish to poll for new email? Example: /var/lib/mail/helpdesk/",
                max_length=200,
                null=True,
                verbose_name="E-Mail Local Directory",
            ),
        ),
        migrations.AddField(
            model_name="queue",
            name="logging_dir",
            field=models.CharField(
                blank=True,
                help_text="If logging is enabled, what directory should we use to store log files for this queue? The standard logging mechanims are used if no directory is set",
                max_length=200,
                null=True,
                verbose_name="Logging Directory",
            ),
        ),
        migrations.AddField(
            model_name="queue",
            name="logging_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("none", "None"),
                    ("debug", "Debug"),
                    ("info", "Information"),
                    ("warn", "Warning"),
                    ("error", "Error"),
                    ("crit", "Critical"),
                ],
                help_text="Set the default logging level. All messages at that level or above will be logged to the directory set below. If no level is set, logging will be disabled.",
                max_length=5,
                null=True,
                verbose_name="Logging Type",
            ),
        ),
        migrations.AlterField(
            model_name="queue",
            name="email_box_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("pop3", "POP 3"),
                    ("imap", "IMAP"),
                    ("local", "Local Directory"),
                ],
                help_text="E-Mail server type for creating tickets automatically from a mailbox - both POP3 and IMAP are supported, as well as reading from a local directory.",
                max_length=5,
                null=True,
                verbose_name="E-Mail Box Type",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("helpdesk", "0003_initial_data_import"),
    ]

    operations = [
        migrations.CreateModel(
            name="QueueMembership",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "queues",
                    models.ManyToManyField(
                        to="helpdesk.Queue", verbose_name="Authorized Queues"
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        verbose_name="User",
                        to=settings.AUTH_USER_MODEL,
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name": "Queue Membership",
                "verbose_name_plural": "Queue Memberships",
            },
            bases=(models.Model,),
        ),
    ]



# Generated by Django 2.0.5 on 2019-03-07 20:30

from django.db import migrations, models
import django.db.models.deletion
import helpdesk.models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0025_queue_dedicated_time"),
    ]

    operations = [
        migrations.CreateModel(
            name="KBIAttachment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "file",
                    models.FileField(
                        max_length=1000,
                        upload_to=helpdesk.models.attachment_path,
                        verbose_name="File",
                    ),
                ),
                (
                    "filename",
                    models.CharField(max_length=1000, verbose_name="Filename"),
                ),
                (
                    "mime_type",
                    models.CharField(max_length=255, verbose_name="MIME Type"),
                ),
                (
                    "size",
                    models.IntegerField(
                        help_text="Size of this file in bytes", verbose_name="Size"
                    ),
                ),
                (
                    "kbitem",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="helpdesk.KBItem",
                        verbose_name="Knowledge base item",
                    ),
                ),
            ],
            options={
                "verbose_name": "Attachment",
                "verbose_name_plural": "Attachments",
                "ordering": ("filename",),
                "abstract": False,
            },
        ),
        migrations.RenameModel(
            old_name="Attachment",
            new_name="FollowUpAttachment",
        ),
    ]



# Generated by Django 3.1.13 on 2021-08-18 14:30

from django.db import migrations, models
import helpdesk.models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0034_create_email_template_for_merged"),
    ]

    operations = [
        migrations.AlterField(
            model_name="usersettings",
            name="email_on_ticket_change",
            field=models.BooleanField(
                default=helpdesk.models.email_on_ticket_change_default,
                help_text="If you're the ticket owner and the ticket is changed via the web by somebody else,do you want to receive an e-mail?",
                verbose_name="E-mail me on ticket change?",
            ),
        ),
    ]



class IgnoreTicketException(Exception):
    """
    Raised when an email message is received from a sender who is marked to be ignored
    """

    pass


class DeleteIgnoredTicketException(Exception):
    """
    Raised when an email message is received from a sender who is marked to be ignored
    and the record is tagged to delete the email from the inbox
    """

    pass



# Generated by Django 4.2 on 2023-04-28 21:23

from django.db import migrations, models
import django.db.models.deletion
import helpdesk.models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0037_alter_queue_email_box_type"),
    ]

    operations = [
        migrations.CreateModel(
            name="Checklist",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="Name")),
                (
                    "ticket",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="checklists",
                        to="helpdesk.ticket",
                        verbose_name="Ticket",
                    ),
                ),
            ],
            options={
                "verbose_name": "Checklist",
                "verbose_name_plural": "Checklists",
            },
        ),
        migrations.CreateModel(
            name="ChecklistTemplate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="Name")),
                (
                    "task_list",
                    models.JSONField(
                        validators=[helpdesk.models.is_a_list_without_empty_element],
                        verbose_name="Task List",
                    ),
                ),
            ],
            options={
                "verbose_name": "Checklist Template",
                "verbose_name_plural": "Checklist Templates",
            },
        ),
        migrations.CreateModel(
            name="ChecklistTask",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "description",
                    models.CharField(max_length=250, verbose_name="Description"),
                ),
                (
                    "completion_date",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="Completion Date"
                    ),
                ),
                (
                    "position",
                    models.PositiveSmallIntegerField(
                        db_index=True, verbose_name="Position"
                    ),
                ),
                (
                    "checklist",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tasks",
                        to="helpdesk.checklist",
                        verbose_name="Checklist",
                    ),
                ),
            ],
            options={
                "verbose_name": "Checklist Task",
                "verbose_name_plural": "Checklist Tasks",
                "ordering": ("position",),
            },
        ),
    ]



#!/usr/bin/python
"""
django-helpdesk - A Django powered ticket tracker for small enterprise.

(c) Copyright 2008 Jutda. All Rights Reserved. See LICENSE for details.

scripts/escalate_tickets.py - Easy way to escalate tickets based on their age,
                              designed to be run from Cron or similar.
"""

from datetime import date, timedelta
from django.core.management.base import BaseCommand
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext as _
from helpdesk.lib import safe_template_context
from helpdesk.models import EscalationExclusion, Queue, Ticket


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "-q",
            "--queues",
            nargs="*",
            choices=list(Queue.objects.values_list("slug", flat=True)),
            help="Queues to include (default: all). Enter the queues slug as space separated list.",
        )
        parser.add_argument(
            "-x",
            "--escalate-verbosely",
            action="store_true",
            default=False,
            help="Display escalated tickets",
        )
        parser.add_argument(
            "-n",
            "--notify-only",
            action="store_true",
            default=False,
            help="Send email reminder but dont escalate tickets",
        )

    def handle(self, *args, **options):
        verbose = options["escalate_verbosely"]
        notify_only = options["notify_only"]

        queue_slugs = options["queues"]
        # Only include queues with escalation configured
        queues = Queue.objects.filter(escalate_days__isnull=False).exclude(
            escalate_days=0
        )
        if queue_slugs is not None:
            queues = queues.filter(slug__in=queue_slugs)

        if verbose:
            self.stdout.write(f"Processing: {queues}")

        for queue in queues:
            last = date.today() - timedelta(days=queue.escalate_days)
            today = date.today()
            workdate = last

            days = 0

            while workdate < today:
                if not EscalationExclusion.objects.filter(date=workdate).exists():
                    days += 1
                workdate = workdate + timedelta(days=1)

            req_last_escl_date = timezone.now() - timedelta(days=days)

            for ticket in (
                queue.ticket_set.filter(status__in=Ticket.OPEN_STATUSES)
                .exclude(priority=1)
                .filter(Q(on_hold__isnull=True) | Q(on_hold=False))
                .filter(
                    Q(last_escalation__lte=req_last_escl_date)
                    | Q(last_escalation__isnull=True, created__lte=req_last_escl_date)
                )
            ):
                ticket.last_escalation = timezone.now()
                ticket.priority -= 1
                ticket.save()

                context = safe_template_context(ticket)

                ticket.send(
                    {
                        "submitter": ("escalated_submitter", context),
                        "ticket_cc": ("escalated_cc", context),
                        "assigned_to": ("escalated_owner", context),
                    },
                    fail_silently=True,
                )

                if verbose:
                    self.stdout.write(
                        f"  - Esclating {ticket.ticket} from {ticket.priority + 1}>{ticket.priority}"
                    )

                if not notify_only:
                    followup = ticket.followup_set.create(
                        title=_("Ticket Escalated"),
                        public=True,
                        comment=_("Ticket escalated after %(nb)s days")
                        % {"nb": queue.escalate_days},
                    )

                    followup.ticketchange_set.create(
                        field=_("Priority"),
                        old_value=ticket.priority + 1,
                        new_value=ticket.priority,
                    )



# Generated by Django 5.2.1 on 2025-06-02 09:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("helpdesk", "0038_checklist_checklisttemplate_checklisttask"),
    ]

    operations = [
        migrations.AlterField(
            model_name="ticketchange",
            name="field",
            field=models.CharField(max_length=200, verbose_name="Field"),
        ),
    ]
